"""
The fusion strategy. The goal of the fusion strategy is to find a pair of
adjacent rows, or adjacent columns such that they can be viewed as a single
column, with a line drawn between them. When this fusion happens, an assumption
that we can count the number of points in the fused row or column is added.

When we map an assumption from the parent tiling to the fused tiling, if it
uses either of the rows or columns being fused, then it is mapped to cover the
entire fused region. With this in mind, we assume that a tiling cannot fuse if
it has an assumption that intersects only partially with one of the adjacent
rows or columns.

We will assume we are always fusing two adjacent columns, and discuss the left
and right hand sides accordingly.
"""
import enum
from collections import Counter, defaultdict
from functools import reduce
from operator import mul
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple

import sympy

from comb_spec_searcher import Constructor
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling

__all__ = ["FusionConstructor"]


class FusionConstructor(Constructor[Tiling, GriddedPerm]):
    """
    The fusion constructor. It will multiply by (fuse_paramater + 1), and
    otherwise pass on the variables.

    - fuse_parameter:           parameter corresponding to the region of the
                                tiling of the child where a line must be drawn.
    - extra_parameters:         a dictionary where the keys are each of the
                                parent parameters pointing to the child
                                parameter it was mapped to. Note, if [ A | A ]
                                fuses to [ A ] then we assume any one sided
                                variable maps to the [ A ] on the child.
    - left_sided_parameters:    all of the parent parameters which overlap
                                fully the left side of the region that is
                                being fused.
    - right_sided_parameters:   all of the parent parameters which overlap
                                fully the right side of the region that is
                                being fused.
    - both_sided_parameters:    all of the parent parameters which overlap
                                fully the entire region that is being fused
    """

    # pylint: disable=too-many-instance-attributes
    # This pylint warning is ignored due to adding in a new algorithm for
    # computing terms that does not apply to sampling. Therefore, this class
    # has two fundamentally different approaches to counting and we would
    # need to refactor the old one to pass this test.
    def __init__(
        self,
        parent: Tiling,
        child: Tiling,
        fuse_parameter: str,
        extra_parameters: Dict[str, str],
        left_sided_parameters: Iterable[str],
        right_sided_parameters: Iterable[str],
        both_sided_parameters: Iterable[str],
        min_left: int,
        min_right: int,
    ):
        # parent -> child parameters
        self.extra_parameters = extra_parameters
        # the reverse of the above list, although different parent parameters could
        # point to the same child parameter as when an assumption in the fused region
        # is fused, we map it to the row or column if it uses at least one row
        # or column in the fusion region.
        self.reversed_extra_parameters: Dict[str, List[str]] = defaultdict(list)
        for parent_var, child_var in self.extra_parameters.items():
            self.reversed_extra_parameters[child_var].append(parent_var)
        # the child parameter that determine 'where the line is drawn'.
        self.fuse_parameter = fuse_parameter
        # sets to tell if parent assumption is one sided, or both
        self.left_sided_parameters = frozenset(left_sided_parameters)
        self.right_sided_parameters = frozenset(right_sided_parameters)
        self.both_sided_parameters = frozenset(both_sided_parameters)

        self._init_checked()

        self.parent_fusion_parameters = self.reversed_extra_parameters[
            self.fuse_parameter
        ]
        self.fusion_types = [
            (
                "left"
                if parent_fusion_parameter in self.left_sided_parameters
                else "right"
                if parent_fusion_parameter in self.right_sided_parameters
                else "both"
            )
            for parent_fusion_parameter in self.parent_fusion_parameters
        ]
        self.predeterminable_left_right_points = [
            parent_vars
            for child_var, parent_vars in self.reversed_extra_parameters.items()
            if child_var != self.fuse_parameter and len(parent_vars) >= 2
        ]
        self.min_points = min_left, min_right

        index_mapping = {
            child.extra_parameters.index(child_param): tuple(
                map(parent.extra_parameters.index, parent_params)
            )
            for child_param, parent_params in self.reversed_extra_parameters.items()
        }
        self.left_parameter_indices = tuple(
            i
            for i, k in enumerate(parent.extra_parameters)
            if k in self.left_sided_parameters
        )
        self.right_parameter_indices = tuple(
            i
            for i, k in enumerate(parent.extra_parameters)
            if k in self.right_sided_parameters
        )
        self.fuse_parameter_index = child.extra_parameters.index(self.fuse_parameter)
        child_pos_to_parent_pos = tuple(
            index_mapping[idx] for idx in range(len(child.extra_parameters))
        )
        self.children_param_map = self.build_param_map(
            child_pos_to_parent_pos, len(parent.extra_parameters)
        )

    def _init_checked(self):
        """
        The lists in reversed_extra_parameters can have size at most three.
        This is checked in the first assertion.

        Moreover, if two parent assumptions map to the same assumption, then one
        of them is one sided, and the other covers both sides, OR they are all
        contained fully in fuse region. This is checked in the second assertion.
        """
        assert all(len(val) <= 3 for val in self.reversed_extra_parameters.values())
        assert all(
            (
                any(
                    parent_var in self.left_sided_parameters
                    or parent_var in self.right_sided_parameters
                    for parent_var in parent_vars
                )
                and any(
                    parent_var in self.both_sided_parameters
                    for parent_var in parent_vars
                )
            )
            or all(
                self.extra_parameters[parent_var] == self.fuse_parameter
                for parent_var in parent_vars
            )
            for parent_vars in self.reversed_extra_parameters.values()
            if len(parent_vars) == 2
        )

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        if max(self.min_points) > 1:
            raise NotImplementedError(
                "not implemented equation in the case of "
                "left or right containing more than one point"
            )
        rhs_func = rhs_funcs[0]
        subs: Dict[str, Any] = {
            child: reduce(mul, [sympy.var(k) for k in parent_vars], 1)
            for child, parent_vars in self.reversed_extra_parameters.items()
        }
        left_vars = reduce(
            mul,
            [
                sympy.var(k)
                for k in self.left_sided_parameters
                if k not in self.parent_fusion_parameters
            ],
            1,
        )
        right_vars = reduce(
            mul,
            [
                sympy.var(k)
                for k in self.right_sided_parameters
                if k not in self.parent_fusion_parameters
            ],
            1,
        )
        p, q = sympy.Number(1), sympy.Number(1)
        for parent_fuse_parameter, fuse_type in zip(
            self.parent_fusion_parameters, self.fusion_types
        ):
            if fuse_type in ("left", "both"):
                p *= sympy.var(parent_fuse_parameter)
            if fuse_type in ("right", "both"):
                q *= sympy.var(parent_fuse_parameter)
        if left_vars == 1 and right_vars == 1 and p == q:
            raise NotImplementedError(
                "Not handled case with no left and right vars, and new fuse "
                "parameter, or only parent fusion parameter covered entire region"
            )
        subs1 = {**subs}
        subs1[self.fuse_parameter] = q / left_vars
        subs2 = {**subs}
        subs2[self.fuse_parameter] = p / right_vars

        left_right_empty = (
            rhs_func.subs(subs2, simultaneous=True),
            rhs_func.subs(subs1, simultaneous=True),
        )
        to_subtract = 0
        if self.min_points[0] == 1:
            # left side is positive, so the right can't be empty
            to_subtract += left_right_empty[1]
        if self.min_points[1] == 1:
            # right side is positive, so thr left can't be empty
            to_subtract += left_right_empty[0]

        return sympy.Eq(
            lhs_func,
            (
                (q * right_vars * rhs_func.subs(subs1, simultaneous=True))
                - (p * left_vars * rhs_func.subs(subs2, simultaneous=True))
            )
            / (q * right_vars - p * left_vars)
            - to_subtract,
        )

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        """
        Uses the `subterms` functions to and the `children_param_maps` to compute
        the terms of size `n`.
        """
        new_terms: Terms = Counter()

        min_left, min_right = self.min_points

        def add_new_term(
            params: List[int], value: int, left_points: int, fuse_region_points: int
        ) -> None:
            """Update new terms if there is enough points on the left and right."""
            if (
                min_left <= left_points
                and min_right <= fuse_region_points - left_points
            ):
                new_terms[tuple(params)] += value

        for param, value in subterms[0](n).items():
            fuse_region_points = param[self.fuse_parameter_index]
            new_params = list(self.children_param_map(param))
            for idx in self.left_parameter_indices:
                new_params[idx] -= fuse_region_points
            add_new_term(new_params, value, 0, fuse_region_points)
            for left_points in range(1, fuse_region_points + 1):
                for idx in self.left_parameter_indices:
                    new_params[idx] += 1
                for idx in self.right_parameter_indices:
                    new_params[idx] -= 1

                add_new_term(new_params, value, left_points, fuse_region_points)
        return new_terms

    def determine_number_of_points_in_fuse_region(
        self, n: int, **parameters: int
    ) -> Iterator[Tuple[int, int]]:
        """
        There are two cases we use to determine the number of left and right points.

        # Case 1:
        There was an assumption A on the parent which maps precisely to the fused
        region. It must be either on the left, right, or covering both columns,
        crucially fully contained within the region to be fused.

        In this case we can determine:
        - if A is left sided, then we know the number of points on the left must be the
        number of points in A
        - if A is right sided, then we know the number of points on the right must be
        be the number of points in A.
        - if A is both sided, then this tells as the sum of the number of left points
        and right points must be equal to the number of points in A. In particular,
        the number of points in A gives us an upper bound for the number of points
        on the left or the right.

        # Case 2:
        We're not in case 1, however there are two assumptions A and B which are mapped
        to the same region on the fused tiling. This means that one of A or B must use
        just the left or right column. Due to the nature of these regions always
        remaining rectangles, this tells us that the other must use both columns.
        W.l.o.g, we will assume the B is the assumption covering both columns

        In this case we can determine:
        - if A uses the left column, then number of points on the right is the number
        of points in B substract the number of points in A
        - if A uses the right column, then the number of points on the left is the
        number of points in B substract the number of points in A
        - the number of points in the entire region is upper bounded by the number of
        points in B.


        In this case, the fusion_type for each fusion parameter is set as follows:
            - left: the parent region was only the left of the unfused region
            - right: the parent region was only the right of the unfused region
            - both: the parent region was all of the unfused region.
        """
        (
            min_left_points,
            max_left_points,
            min_right_points,
            max_right_points,
            min_both_points,
            max_both_points,
        ) = self._min_max_points_by_fuse_parameters(n, **parameters)

        if (
            min_left_points > max_left_points
            or min_right_points > max_right_points
            or min_both_points > max_both_points
        ):
            return

        for overlapping_parameters in self.predeterminable_left_right_points:
            (
                new_left,
                new_right,
            ) = self._determine_number_of_points_by_overlapping_parameter(
                overlapping_parameters, **parameters
            )
            if new_left is not None:
                min_left_points = max(min_left_points, new_left)
                max_left_points = min(min_left_points, new_left)
            if new_right is not None:
                min_right_points = max(min_right_points, new_right)
                max_right_points = min(max_right_points, new_right)
            if min_left_points > max_left_points or min_right_points > max_right_points:
                return

        min_both_points = max(min_both_points, min_left_points + min_right_points)
        max_both_points = min(max_both_points, max_right_points + max_left_points)
        for number_left_points in range(min_left_points, max_left_points + 1):
            for number_right_points in range(min_right_points, max_right_points + 1):
                both = number_left_points + number_right_points
                if both < min_both_points:
                    continue
                if both > max_both_points:
                    break
                yield number_left_points, number_right_points

    def _min_max_points_by_fuse_parameters(
        self, n: int, **parameters: int
    ) -> Tuple[int, int, int, int, int, int]:
        """
        # Case 1:
        There was an assumption A on the parent which maps precisely to the fused
        region. It must be either on the left, right, or covering both columns,
        crucially fully contained within the region to be fused.

        In this case we can determine:
        - if A is left sided, then we know the number of points on the left must be the
        number of points in A
        - if A is right sided, then we know the number of points on the right must be
        be the number of points in A.
        - if A is both sided, then this tells as the sum of the number of left points
        and right points must be equal to the number of points in A. In particular,
        the number of points in A gives us an upper bound for the number of points
        on the left or the right.
        """
        min_left_points, max_left_points = self.min_points[0], n - self.min_points[1]
        min_right_points, max_right_points = self.min_points[1], n - self.min_points[0]
        min_both_points, max_both_points = sum(self.min_points), n
        for parent_fusion_parameter, fusion_type in zip(
            self.parent_fusion_parameters, self.fusion_types
        ):
            number_points_parent_fuse_parameter = parameters[parent_fusion_parameter]
            if fusion_type == "left":
                min_left_points = max(
                    min_left_points, number_points_parent_fuse_parameter
                )
                max_left_points = min(
                    max_left_points, number_points_parent_fuse_parameter
                )
            elif fusion_type == "right":
                min_right_points = max(
                    min_right_points, number_points_parent_fuse_parameter
                )
                max_right_points = min(
                    max_right_points, number_points_parent_fuse_parameter
                )
            else:
                assert fusion_type == "both"
                # TODO: is this the right way?
                max_left_points = min(
                    max_left_points, number_points_parent_fuse_parameter
                )
                max_right_points = min(
                    max_right_points, number_points_parent_fuse_parameter
                )
                min_both_points = max(
                    min_both_points, number_points_parent_fuse_parameter
                )
                max_both_points = min(
                    max_both_points, number_points_parent_fuse_parameter
                )
            if (
                min_left_points > max_left_points
                or min_right_points > max_right_points
                or min_both_points > max_both_points
            ):
                break

        return (
            min_left_points,
            max_left_points,
            min_right_points,
            max_right_points,
            min_both_points,
            max_both_points,
        )

    def _determine_number_of_points_by_overlapping_parameter(
        self, overlapping_parameters: List[str], **parameters: int
    ) -> Tuple[Optional[int], Optional[int]]:
        """
        # Case 2:
        There are two assumptions A and B which are mapped to the same region
        on the fused tiling (which is not the fused region, else we'd be in case
        1). This means that one of A or B must use just the left or right column.
        Due to the nature of these regions always remaining rectangles, this tells
        us that the other must use both columns.
        W.l.o.g, we will assume the B is the assumption covering both columns

        In this case we can determine:
        - if A uses the left column, then number of points on the right is the number
        of points in B substract the number of points in A
        - if A uses the right column, then the number of points on the left is the
        number of points in B substract the number of points in A
        - the number of points in the entire region is upper bounded by the number of
        points in B.
        """
        p1 = overlapping_parameters[0]
        p2 = overlapping_parameters[1]
        p1_left = p1 in self.left_sided_parameters
        p1_right = p1 in self.right_sided_parameters
        p2_left = p2 in self.left_sided_parameters
        p2_right = p2 in self.right_sided_parameters
        # TODO: tidy up this function, and update doc string
        assert not (p1_left and p2_right) and not (p1_right and p2_left)
        if p1_left:
            assert p2 in self.both_sided_parameters
            return None, parameters[p2] - parameters[p1]
        if p1_right:
            assert p2 in self.both_sided_parameters
            return parameters[p2] - parameters[p1], None
        if p2_left:
            assert p1 in self.both_sided_parameters
            return None, parameters[p1] - parameters[p2]
        if p2_right:
            assert p1 in self.both_sided_parameters
            return parameters[p1] - parameters[p2], None
        raise ValueError("Overlapping parameters overlap same region")

    def update_subparams(
        self, number_of_left_points: int, number_of_right_points: int, **parameters: int
    ) -> Optional[Dict[str, int]]:
        """
        Return the updates dictionary of parameters, such that each parameter points
        to the child parameter.

        The extra parameters mapping may not be unique, so if
        two updated parameters have a different value, then the function returns None
        to tell the calling function that the value of the subrec call should be 0.

        Also, number_of_left_points is added to the parameter if it is unnfused
        to include only the right side of the fused region, and number of right
        points is added to the region is it is unfused to include only the left
        side.
        """
        res = {self.fuse_parameter: number_of_left_points + number_of_right_points}
        for parameter, value in parameters.items():
            if parameter not in self.extra_parameters and value != 0:
                return None
            if (
                parameter in self.left_sided_parameters
                and number_of_left_points > value
            ) or (
                parameter in self.right_sided_parameters
                and number_of_right_points > value
            ):
                return None
            updated_value = (
                value + number_of_left_points
                if parameter in self.right_sided_parameters
                else value + number_of_right_points
                if parameter in self.left_sided_parameters
                else value
            )
            child_parameter = self.extra_parameters[parameter]
            if child_parameter not in res:
                res[child_parameter] = updated_value
            elif updated_value != res[child_parameter]:
                return None
        return res

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError(
            "This is implemented on the FusionRule class directly"
        )

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError(
            "This is implemented on the FusionRule class directly"
        )

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        # Base cases (instance and count checks)
        init = self._equiv_base_cases(other, data)
        if init is None:
            return False, None

        # p1 and p2 are parent parameters and n their length, rest is empty
        p1, p2, n, bijection, in_use, term_list, term_funcs = init

        # Find a bijection between p1 and p2 that is consistent
        def _backtrack(rev: bool):
            assert isinstance(other, type(self))
            for x in range(n):
                if x in in_use:
                    continue

                bijection.append(x)
                in_use.add(x)

                last = len(bijection) == n

                # If consistent and either done or recursively successful
                result = self._equiv_backtrack_consistent(
                    bijection, rev, other, p1, p2
                ) and (last or _backtrack(rev))

                # Check if terms match
                if result and last:
                    if not term_list:
                        for i in range(term_funcs[2]):
                            term_list.append((term_funcs[0](i), term_funcs[1](i)))
                    if not FusionConstructor._term_consistent(bijection, term_list):
                        result = False

                bijection.pop()
                in_use.remove(x)

                # Found
                if result:
                    return True
            return False

        lis: List[bool] = []
        # l-l and r-r match: Data appends False
        if _backtrack(False):
            lis.append(False)
        assert len(in_use) == 0
        # l-r and r-l match: Data appends True
        if _backtrack(True):
            lis.append(True)

        # Data will be [], [True], [False], [False, True]
        return len(lis) > 0, lis

    def _equiv_base_cases(
        self,
        other: "Constructor",
        data: Optional[object],
    ) -> Optional[
        Tuple[
            List[str],
            List[str],
            int,
            List[int],
            Set[int],
            List[Tuple[Terms, Terms]],
            Tuple[Callable[[int], Terms], Callable[[int], Terms], int],
        ]
    ]:
        if not isinstance(other, type(self)):
            return None
        # Early domain failure
        p1 = list(self.extra_parameters.keys())
        p2 = list(other.extra_parameters.keys())
        if len(p1) != len(p2):
            return None
        # Early codomain failure
        vals1 = Counter(self.extra_parameters.values())
        vals2 = Counter(other.extra_parameters.values())
        if (
            len(vals1) != len(vals2)
            or sorted(vals1.values()) != sorted(vals2.values())
            or (self.fuse_parameter in vals1) != (other.fuse_parameter in vals2)  # xor
        ):
            return None
        # Extract typed data from optional object
        if data is None:
            raise ValueError("Terms are needed to compare fusion constructors")
        assert isinstance(data, tuple) and len(data) == 3
        f1: Callable[[int], Terms] = data[0]
        f2: Callable[[int], Terms] = data[1]
        max_check: int = data[2]
        term_lis: List[Tuple[Terms, Terms]] = []
        return p1, p2, len(p1), [], set(), term_lis, (f1, f2, max_check)

    def _equiv_backtrack_consistent(
        self,
        bi: List[int],
        rev: bool,
        other: "FusionConstructor",
        p1: List[str],
        p2: List[str],
    ):
        grp_left, grp_right = set(), set()
        for i, j in enumerate(bi):
            # If one, of matched parameters, is in both sides in one but not the other
            # or if one maps to fuse parameter and the other does not.
            if (p1[i] in self.both_sided_parameters) != (
                p2[j] in other.both_sided_parameters
            ) or (self.extra_parameters[p1[i]] == self.fuse_parameter) != (
                other.extra_parameters[p2[j]] == other.fuse_parameter
            ):
                return False
            # Gather those in p2 that corresponds to each side in p1
            if p1[i] in self.left_sided_parameters:
                grp_left.add(p2[j])
            if p1[i] in self.right_sided_parameters:
                grp_right.add(p2[j])
        # If rev, corresponding parameters of those in one's left should be in other's
        # right. If not rev, they should be on the same side (right-right, left-left).
        if rev:
            return grp_right.issubset(
                other.left_sided_parameters
            ) and grp_left.issubset(other.right_sided_parameters)
        return grp_left.issubset(other.left_sided_parameters) and grp_right.issubset(
            other.right_sided_parameters
        )

    @staticmethod
    def _term_consistent(
        bijection: List[int], terms: List[Tuple[Terms, Terms]]
    ) -> bool:
        """Check if the bijection is valid with respect to terms."""
        for t1, t2 in terms:
            for k, v in t1.items():
                k2 = tuple(k[i] for i in bijection)
                if k2 not in t2 or t2[k2] != v:
                    return False
        return True


class ReverseFusionConstructor(Constructor[Tiling, GriddedPerm]):
    """
    The reverse fusion constructor

    - fuse_parameter:           parameter corresponding to the region of the
                                tiling of the child where a line must be drawn.
    - extra_parameters:         a dictionary where the keys are each of the
                                parent parameters pointing to the child
                                parameter it was mapped to. Note, if [ A | A ]
                                fuses to [ A ] then we assume any one sided
                                variable maps to the [ A ] on the child.
    - left_sided_parameters:    all of the parent parameters which overlap
                                fully the left side of the region that is
                                being fused.
    - right_sided_parameters:   all of the parent parameters which overlap
                                fully the right side of the region that is
                                being fused.
    """

    class Type(enum.Enum):
        LEFT_ONLY = enum.auto()
        RIGHT_ONLY = enum.auto()
        BOTH = enum.auto()

    def __init__(
        self,
        t_unfuse: Tiling,
        t_fuse: Tiling,
        fuse_parameter: str,
        extra_parameters: Dict[str, str],
        left_sided_parameters: Tuple[str, ...],
        right_sided_parameters: Tuple[str, ...],
        left_points: int,
        right_points: int,
    ):
        self.left_points = left_points
        self.right_points = right_points
        left_fuse_index = self.get_left_fuse_index(
            left_sided_parameters, fuse_parameter, extra_parameters, t_unfuse
        )
        right_fuse_index = self.get_left_fuse_index(
            right_sided_parameters, fuse_parameter, extra_parameters, t_unfuse
        )
        if left_fuse_index is not None:
            if right_fuse_index is not None:
                self.type = ReverseFusionConstructor.Type.BOTH
            else:
                self.type = ReverseFusionConstructor.Type.LEFT_ONLY
        else:
            assert right_fuse_index is not None
            self.type = ReverseFusionConstructor.Type.RIGHT_ONLY
        self.unfuse_pos_to_fuse_pos = self.build_unfuse_pos_to_fuse_pos(
            t_unfuse,
            t_fuse,
            extra_parameters,
            left_fuse_index,
            left_sided_parameters,
            right_fuse_index,
            right_sided_parameters,
        )
        self.left_sided_index = tuple(
            map(t_unfuse.extra_parameters.index, left_sided_parameters)
        )
        self.right_sided_index = tuple(
            map(t_unfuse.extra_parameters.index, right_sided_parameters)
        )
        self.num_fuse_param = len(t_fuse.extra_parameters)

    @staticmethod
    def build_unfuse_pos_to_fuse_pos(
        t_unfuse: Tiling,
        t_fuse: Tiling,
        extra_parameters: Dict[str, str],
        left_fuse_param_idx: Optional[int],
        left_sided_parameters: Tuple[str, ...],
        right_fuse_param_idx: Optional[int],
        right_sided_parameters: Tuple[str, ...],
    ) -> Tuple[Tuple[int, ...], ...]:
        """
        Return the contribution of the unfuse assumption to the fused assumptions.

        The i-th tuple gives the position of all the assumptions of the fused tiling
        the i-th assumption of the unfused tiling contributes to.
        """
        assert left_fuse_param_idx is not None or right_fuse_param_idx is not None
        reversed_extra_params: Dict[str, Set[str]] = defaultdict(set)
        for u_param, f_param in extra_parameters.items():
            reversed_extra_params[f_param].add(u_param)
        res: Tuple[Set[int], ...] = tuple(set() for _ in t_unfuse.extra_parameters)
        one_sided_parameters = set(left_sided_parameters + right_sided_parameters)
        for f_param, u_params in reversed_extra_params.items():
            f_param_idx = t_fuse.extra_parameters.index(f_param)
            for u_param in u_params:
                u_param_idx = t_unfuse.extra_parameters.index(u_param)
                if u_param not in one_sided_parameters:
                    res[u_param_idx].add(f_param_idx)
                    break
                if (
                    u_param in left_sided_parameters
                    and right_fuse_param_idx is not None
                ):
                    res[u_param_idx].add(f_param_idx)
                    res[right_fuse_param_idx].add(f_param_idx)
                    break
                if (
                    u_param in right_sided_parameters
                    and left_fuse_param_idx is not None
                ):
                    res[u_param_idx].add(f_param_idx)
                    res[left_fuse_param_idx].add(f_param_idx)
                    break
            else:
                for u_param in u_params:
                    u_param_idx = t_unfuse.extra_parameters.index(u_param)
                    res[u_param_idx].add(f_param_idx)
        return tuple(map(tuple, res))

    @staticmethod
    def get_left_fuse_index(
        left_sided_parameters: Tuple[str, ...],
        fuse_parameter: str,
        extra_parameters: Dict[str, str],
        t_unfuse: Tiling,
    ) -> Optional[int]:
        """
        Return the param index of the left sided param that fuses to the fuse param on
        the fused tiling.
        """
        for parent_param, child_param in extra_parameters.items():
            if child_param == fuse_parameter and parent_param in left_sided_parameters:
                return t_unfuse.extra_parameters.index(parent_param)
        return None

    def forward_map(self, param: Parameters) -> Parameters:
        """
        Maps a set of parameters on the fuse tiling to a set of parameters on
        the unfused tiling.
        """
        new_param = [0 for _ in range(self.num_fuse_param)]
        assert len(param) == len(self.unfuse_pos_to_fuse_pos)
        for pvalue, fuse_idxs in zip(param, self.unfuse_pos_to_fuse_pos):
            for fuse_idx in fuse_idxs:
                new_param[fuse_idx] += pvalue
                if (
                    self.type == ReverseFusionConstructor.Type.LEFT_ONLY
                    and self.right_points
                    and fuse_idx in self.left_sided_index
                ):
                    new_param[fuse_idx] += 1
                elif (
                    self.type == ReverseFusionConstructor.Type.RIGHT_ONLY
                    and self.left_points
                    and fuse_idx in self.right_sided_index
                ):
                    new_param[fuse_idx] += 1
        return tuple(new_param)

    def a_map(self, param: Parameters) -> Parameters:
        if self.type == ReverseFusionConstructor.Type.BOTH:
            # In the case where both is track we don't subtract anything
            return tuple(-1 for _ in param)
        if self.type == ReverseFusionConstructor.Type.LEFT_ONLY:
            add = self.left_sided_index
            substract = self.right_sided_index
        else:
            assert self.type == ReverseFusionConstructor.Type.RIGHT_ONLY
            add = self.right_sided_index
            substract = self.left_sided_index
        new_param = list(param)
        for idx in add:
            new_param[idx] += 1
        for idx in substract:
            new_param[idx] -= 1
        return tuple(new_param)

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        terms: Terms = Counter()
        child_terms = subterms[0](n)
        for param, value in child_terms.items():
            new_param = self.forward_map(param)
            new_value = value - child_terms[self.a_map(param)]
            assert new_value >= 0
            if new_value > 0:
                assert new_param not in terms or new_value == terms[new_param]
                terms[new_param] = new_value
        return terms

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        raise NotImplementedError

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError("Required for bijections")



import abc
from itertools import chain
from typing import Dict, Iterator, Optional, Tuple

from comb_spec_searcher import Constructor
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import (
    DisjointUnionStrategy,
    Strategy,
    StrategyFactory,
)
from tilings import GriddedPerm, Tiling, TrackingAssumption
from tilings.algorithms import Fusion

Cell = Tuple[int, int]


class AssumptionOrPointJumpingStrategy(Strategy[Tiling, GriddedPerm]):
    """
    An abstract strategy class which moves requirements or assumptions from a
    column (or row) to its neighbouring column (or row) if the two columns
    are fusable.
    """

    def __init__(self, idx1: int, idx2: int, row: bool):
        self.idx1 = idx1
        self.idx2 = idx2
        self.row = row
        super().__init__()

    @abc.abstractmethod
    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        pass

    def swapped_requirements(
        self, tiling: Tiling
    ) -> Tuple[Tuple[GriddedPerm, ...], ...]:
        return tuple(tuple(map(self._swapped_gp, req)) for req in tiling.requirements)

    def swapped_assumptions(self, tiling: Tiling) -> Tuple[TrackingAssumption, ...]:
        return tuple(
            ass.__class__(map(self._swapped_gp, ass.gps)) for ass in tiling.assumptions
        )

    def _swapped_gp(self, gp: GriddedPerm) -> GriddedPerm:
        if self._in_both_columns(gp):
            return gp
        return GriddedPerm(gp.patt, map(self._swap_cell, gp.pos))

    def _in_both_columns(self, gp: GriddedPerm) -> bool:
        if self.row:
            return any(y == self.idx1 for _, y in gp.pos) and any(
                y == self.idx2 for _, y in gp.pos
            )
        return any(x == self.idx1 for x, _ in gp.pos) and any(
            x == self.idx2 for x, _ in gp.pos
        )

    def _swap_cell(self, cell: Cell) -> Cell:
        x, y = cell
        if self.row:
            if y == self.idx1:
                y = self.idx2
            elif y == self.idx2:
                y = self.idx1
        else:
            if x == self.idx1:
                x = self.idx2
            elif x == self.idx2:
                x = self.idx1
        return x, y

    def _swap_assumption(self, assumption: TrackingAssumption) -> TrackingAssumption:
        return assumption.__class__(self._swapped_gp(gp) for gp in assumption.gps)

    @abc.abstractmethod
    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        pass

    @abc.abstractmethod
    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm]]:
        pass

    @abc.abstractmethod
    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        pass

    @abc.abstractmethod
    def formal_step(self) -> str:
        pass

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d.pop("ignore_parent")
        d["idx1"] = self.idx1
        d["idx2"] = self.idx2
        d["row"] = self.row
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "AssumptionOrPointJumpingStrategy":
        return cls(d.pop("idx1"), d.pop("idx2"), d.pop("row"))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.idx1}, {self.idx2}, {self.row})"


class AssumptionAndPointJumpingStrategy(
    AssumptionOrPointJumpingStrategy,
    DisjointUnionStrategy[Tiling, GriddedPerm],
):
    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        return (
            Tiling(
                comb_class.obstructions,
                self.swapped_requirements(comb_class),
                self.swapped_assumptions(comb_class),
                simplify=False,
                derive_empty=False,
                remove_empty_rows_and_cols=False,
            ),
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm]]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        child = children[0]
        return (
            {
                comb_class.get_assumption_parameter(
                    ass
                ): child.get_assumption_parameter(self._swap_assumption(ass))
                for ass in comb_class.assumptions
            },
        )

    def formal_step(self) -> str:
        row_or_col = "rows" if self.row else "cols"
        return (
            f"swapping reqs and assumptions in {row_or_col} {self.idx1} and {self.idx2}"
        )


class AssumptionJumpingStrategy(AssumptionOrPointJumpingStrategy):
    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        return (
            Tiling(
                comb_class.obstructions,
                comb_class.requirements,
                self.swapped_assumptions(comb_class),
                simplify=False,
                derive_empty=False,
                remove_empty_rows_and_cols=False,
            ),
        )

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return True

    def is_reversible(self, comb_class: Tiling) -> bool:
        return True

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        return (0,)

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError(
            "Constructor for assumption jumping is not implemented"
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError(
            "Reverse constructor for assumption jumping is not implemented."
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm]]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        raise NotImplementedError(
            "extra_parameters not implemented for assumption jumping"
        )

    def formal_step(self) -> str:
        row_or_col = "rows" if self.row else "cols"
        return f"swapping assumptions in {row_or_col} {self.idx1} and {self.idx2}"


class PointJumpingStrategy(AssumptionOrPointJumpingStrategy):
    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        return (
            Tiling(
                comb_class.obstructions,
                self.swapped_requirements(comb_class),
                comb_class.assumptions,
                simplify=False,
                derive_empty=False,
                remove_empty_rows_and_cols=False,
            ),
        )

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return True

    def is_reversible(self, comb_class: Tiling) -> bool:
        return True

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        return (0,)

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError("Constructor not implemented for point jumping")

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError(
            "Reverse contructor not implemented for point jumping"
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm]]:
        raise NotImplementedError("not implemented map for assumption or point jumping")

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        raise NotImplementedError("not implemented extra_parameters for point jumping")

    def formal_step(self) -> str:
        row_or_col = "rows" if self.row else "cols"
        return f"swapping requirements in {row_or_col} {self.idx1} and {self.idx2}"


class AssumptionAndPointJumpingFactory(StrategyFactory[Tiling]):
    """
    A factory returning the strategies that moves requirements and/or assumptions
    across the boundary of two fusable columns (or rows).
    """

    def __call__(
        self, comb_class: Tiling
    ) -> Iterator[AssumptionOrPointJumpingStrategy]:
        cols, rows = comb_class.dimensions
        gps_to_be_swapped = chain(
            *comb_class.requirements, *[ass.gps for ass in comb_class.assumptions]
        )
        for col in range(cols - 1):
            if any(x in (col, col + 1) for gp in gps_to_be_swapped for x, _ in gp.pos):
                algo = Fusion(comb_class, col_idx=col)
                if algo.fusable():
                    yield AssumptionAndPointJumpingStrategy(col, col + 1, False)
                    yield AssumptionJumpingStrategy(col, col + 1, False)
                    yield PointJumpingStrategy(col, col + 1, False)
        for row in range(rows - 1):
            if any(y in (row, row + 1) for gp in gps_to_be_swapped for y, _ in gp.pos):
                algo = Fusion(comb_class, row_idx=row)
                if algo.fusable():
                    yield AssumptionAndPointJumpingStrategy(row, row + 1, True)
                    yield AssumptionJumpingStrategy(row, row + 1, True)
                    yield PointJumpingStrategy(row, row + 1, True)

    def __str__(self) -> str:
        return "assumption and point jumping"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other: object) -> bool:
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __hash__(self) -> int:
        return hash(self.__class__)

    @classmethod
    def from_dict(cls, d: dict) -> "AssumptionAndPointJumpingFactory":
        assert not d
        return cls()



"""The cell reduction strategy."""
from typing import Callable, Dict, Iterator, List, Optional, Set, Tuple, cast

import sympy

from comb_spec_searcher import Constructor, Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.assumptions import ComponentAssumption, TrackingAssumption

Cell = Tuple[int, int]


class CellReductionConstructor(Constructor):
    def __init__(self, parameter: str):
        self.parameter = parameter

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        raise NotImplementedError

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        raise NotImplementedError

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]]]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[GriddedPerm]:
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError


class CellReductionStrategy(Strategy[Tiling, GriddedPerm]):
    """A strategy that replaces the cell with an increasing or decreasing cell."""

    def __init__(
        self,
        cell: Cell,
        increasing: bool,
        tracked: bool = True,
    ):
        self.cell = cell
        self.tracked = tracked
        self.increasing = increasing
        super().__init__()

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return False

    def is_reversible(self, comb_class: Tiling) -> bool:
        return False

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[int, ...]:
        return (0, 0)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, Tiling]:
        if self.increasing:
            extra = Perm((1, 0))
        else:
            extra = Perm((0, 1))
        reduced_obs = sorted(
            [
                ob
                for ob in comb_class.obstructions
                if not ob.pos[0] == self.cell or not ob.is_single_cell()
            ]
            + [GriddedPerm.single_cell(extra, self.cell)]
        )
        reduced_reqs = sorted(
            req
            for req in comb_class.requirements
            if not all(gp.pos[0] == self.cell and gp.is_single_cell() for gp in req)
        )
        reduced_ass = sorted(
            ass
            for ass in comb_class.assumptions
            if not isinstance(ass, ComponentAssumption)
            or GriddedPerm.point_perm(self.cell) not in ass.gps
        )
        reduced_tiling = Tiling(
            reduced_obs,
            reduced_reqs,
            reduced_ass,
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        local_basis = comb_class.sub_tiling([self.cell])
        if self.tracked:
            return (
                reduced_tiling.add_assumption(
                    TrackingAssumption.from_cells([self.cell])
                ),
                local_basis,
            )
        return reduced_tiling, local_basis

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> CellReductionConstructor:
        if not self.tracked:
            raise NotImplementedError("The reduction strategy was not tracked")
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't reduce the cell")
        ass = TrackingAssumption.from_cells([self.cell])
        child_param = children[0].get_assumption_parameter(ass)
        gp = GriddedPerm.point_perm(self.cell)
        if any(gp in assumption.gps for assumption in comb_class.assumptions):
            raise NotImplementedError
        return CellReductionConstructor(child_param)

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't reduce the cell")
        raise NotImplementedError

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str]]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        raise NotImplementedError

    def formal_step(self) -> str:
        return f"reducing cell {self.cell}"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d.pop("ignore_parent")
        d.pop("inferrable")
        d.pop("possibly_empty")
        d.pop("workable")
        d["cell"] = self.cell
        d["tracked"] = self.tracked
        d["increasing"] = self.increasing
        return d

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"cell={self.cell!r}",
                f"increasing={self.increasing!r}",
                f"tracked={self.tracked!r}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    @classmethod
    def from_dict(cls, d: dict) -> "CellReductionStrategy":
        cell = cast(Tuple[int, int], tuple(d.pop("cell")))
        tracked = d.pop("tracked")
        increasing = d.pop("increasing")
        assert not d
        return cls(cell, increasing, tracked)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"


class CellReductionFactory(StrategyFactory[Tiling]):
    def __init__(self, tracked: bool):
        self.tracked = tracked
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[CellReductionStrategy]:
        if comb_class.dimensions == (1, 1):
            return
        cell_bases = comb_class.cell_basis()
        for cell in self.reducible_cells(comb_class):
            if not (  # a finite cell
                any(patt.is_increasing() for patt in cell_bases[cell][0])
                and any(patt.is_decreasing() for patt in cell_bases[cell][0])
            ) and self.can_reduce_cell_with_requirements_and_assumptions(
                comb_class, cell
            ):
                yield CellReductionStrategy(cell, True, self.tracked)
                yield CellReductionStrategy(cell, False, self.tracked)

    def can_reduce_cell_with_requirements_and_assumptions(
        self, tiling: Tiling, cell: Cell
    ) -> bool:
        return all(  # local
            all(gp.pos[0] == cell and gp.is_single_cell() for gp in req)
            or all(
                # at most one point in cell
                sum(1 for _ in gp.points_in_cell(cell)) < 2
                # no gp in row and col
                and not self.gp_in_row_and_col(gp, cell)
                for gp in req
            )
            for req in tiling.requirements
        ) and all(
            not isinstance(ass, ComponentAssumption)
            or GriddedPerm.point_perm(cell) not in ass.gps
            or len(ass.gps) == 1
            for ass in tiling.assumptions
        )

    @staticmethod
    def gp_in_row_and_col(gp: GriddedPerm, cell: Cell) -> bool:
        """Return True if there are points touching a cell in the row and col of
        cell that isn't cell itself."""
        x, y = cell
        return (
            len(set(gp.pos[idx][1] for idx, _ in gp.get_points_col(x))) > 1
            and len(set(gp.pos[idx][0] for idx, _ in gp.get_points_row(y))) > 1
        )

    def reducible_cells(self, tiling: Tiling) -> Set[Cell]:
        """Return the set of non-monotone cells with at most one point in a crossing
        obstrution touching them, and no obstruction touching a cell in the row and
        a cell in the column."""
        cells = set(tiling.active_cells) - set(tiling.point_cells)
        for gp in tiling.obstructions:
            if not cells:
                break
            if not gp.is_localized():
                seen = set()
                for cell in gp.pos:
                    if cell in seen or self.gp_in_row_and_col(gp, cell):
                        cells.discard(cell)
                    seen.add(cell)
            elif len(gp) == 2:
                cells.discard(gp.pos[0])
        return cells

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.tracked})"

    def __str__(self) -> str:
        return ("tracked " if self.tracked else "") + "cell reduction factory"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "CellReductionFactory":
        return cls(d["tracked"])



from collections import defaultdict
from functools import partial
from itertools import chain, islice
from typing import Callable, Dict, Iterable, List, Optional, Set, Tuple

from ..griddedperm import GriddedPerm
from .minimal_gridded_perms import MinimalGriddedPerms

Cell = Tuple[int, int]
Requirement = Tuple[GriddedPerm, ...]


class GriddedPermReduction:
    def __init__(
        self,
        obstructions: Tuple[GriddedPerm, ...],
        requirements: Tuple[Tuple[GriddedPerm, ...], ...],
        sorted_input: bool = False,
        already_minimized_obs: bool = False,
        manual: bool = False,
    ):
        # Only using MGP for typing purposes.
        if sorted_input:
            self._obstructions = obstructions
            self._requirements = requirements
        else:
            self._obstructions = tuple(sorted(obstructions))
            self._requirements = tuple(sorted(tuple(sorted(r)) for r in requirements))
        if not manual:
            self._minimize_griddedperms(already_minimized_obs=already_minimized_obs)

    @property
    def obstructions(self) -> Tuple[GriddedPerm, ...]:
        return self._obstructions

    @property
    def requirements(self) -> Tuple[Requirement, ...]:
        return self._requirements

    def _minimize_griddedperms(self, already_minimized_obs=False) -> None:
        """Minimizes the set of obstructions and the set of requirement lists.
        The set of obstructions are first reduced to a minimal set. The
        requirements that contain any obstructions are removed from their
        respective lists. If any requirement list is empty, then the tiling is
        empty.
        """

        def set_empty():
            self._obstructions = (GriddedPerm.empty_perm(),)
            self._requirements = tuple()

        if not already_minimized_obs:
            self._obstructions = tuple(
                sorted(GriddedPermReduction._minimize(self._obstructions))
            )
        first = True
        while True:
            # Minimize the set of requiriments
            if not self.minimal_reqs() and not first:
                # if it didn't change then break if its the second or
                # later pass
                break
            first = False

            if self._requirements and not self._requirements[0]:
                set_empty()
                break

            # Minimize the set of obstructions
            if not self.minimal_obs():
                # if they didn't change then break
                break

            if self._obstructions and not self._obstructions[0]:
                set_empty()
                break

    def clean_isolated(
        self, obstructions: Tuple[GriddedPerm, ...], gp: GriddedPerm
    ) -> Set[GriddedPerm]:
        """
        Return the obstructions which have some factors removed due to containing gp.
        """
        cleaned_obs: Set[GriddedPerm] = set()
        for ob in obstructions:
            cells_to_remove: Set[Cell] = set()
            for factor in ob.factors():
                if self._griddedperm_implied_by_requirement(factor, (gp,)):
                    cells_to_remove.update(factor.pos)
            if cells_to_remove:
                cleaned_obs.add(ob.remove_cells(cells_to_remove))
        return cleaned_obs

    def minimal_obs(self) -> bool:
        """
        Reduce the obstruction according to the requirements.
        Return True if something changed.
        """
        if not self._obstructions:
            return False
        changed = False
        new_obs: Set[GriddedPerm] = set()
        for requirement in self.requirements:
            cleaned_obs = tuple(
                tuple(
                    GriddedPermReduction._minimize(
                        self.clean_isolated(self._obstructions, gp)
                    )
                )
                for gp in requirement
            )
            gpr = GriddedPermReduction(self._obstructions, cleaned_obs, manual=True)
            gpr.minimal_reqs()
            cleaned_obs = gpr.requirements
            implied_obs = set(
                MinimalGriddedPerms(
                    self._obstructions,
                    cleaned_obs,
                ).minimal_gridded_perms(
                    max_length_to_build=max(map(len, self._obstructions)) - 1
                )
            )
            new_obs.update(implied_obs)
        if new_obs:
            changed = True
            self._obstructions = tuple(
                sorted(
                    GriddedPermReduction._minimize(self._obstructions + tuple(new_obs))
                )
            )
            self.minimal_obs()
        return changed

    def minimal_reqs(self) -> bool:
        """
        Update the requirements according to avoiding the obstructions.
        Returns True if something changed.
        """
        algos: Tuple[Callable, ...] = (
            GriddedPermReduction.factored_reqs,
            self.remove_redundant,
            self.cleaned_requirements,
            partial(self.remove_avoided, obstructions=self._obstructions),
        )
        minimized_requirements = list(self._requirements)
        for algo in algos:
            minimized_requirements = algo(minimized_requirements)
            if any(not r for r in minimized_requirements):
                self._requirements = (tuple(),)
                return True
        sorted_minimized_requirements = tuple(
            sorted(tuple(sorted(req)) for req in minimized_requirements)
        )
        if sorted_minimized_requirements == self._requirements:
            return False
        self._requirements = sorted_minimized_requirements
        self.minimal_reqs()
        return True

    def remove_redundant(self, requirements: List[Requirement]) -> List[Requirement]:
        relevant_reqs: List[Requirement] = []
        for idx, requirement in enumerate(requirements):
            if not all(requirement):
                continue
            if self._requirement_implied_by_some_requirement(
                requirement,
                chain(
                    islice(requirements, idx),
                    (
                        gps
                        for gps in islice(requirements, idx + 1, None)
                        if sorted(gps) != sorted(requirement)
                    ),
                ),
            ):
                # we only keep requirements which are not implies by other
                # requirements
                continue
            relevant_reqs.append(requirement)
        return relevant_reqs

    # If all of the gp's in a requirment list contain the same factor, you can remove
    # that factor from all of them and add it as it's own size one requirement list
    # [size one requirement list inferral]
    @staticmethod
    def factored_reqs(requirements: Iterable[Requirement]) -> List[Requirement]:
        """
        Add factors of requirements as size one requirements, removing them
        from each of the gridded permutations in the requirement it is
        contained in.
        """
        res: List[Requirement] = []
        for requirement in requirements:
            if not requirement:
                # If req is empty, then this requirement can't be satisfied, so
                # return it and mark obstructions and requirements as empty.
                return [tuple()]
            # If any gridded permutation in list is empty then you vacuously
            # contain this requirement
            if all(requirement):
                factors = GriddedPermReduction.factors(requirement)
                if not factors or (len(factors) == 1 and len(requirement) == 1):
                    # if there are no factors in the intersection, or it is just
                    # the same req as the first, we do nothing and add the original
                    res.append(requirement)
                    continue
                # add each of the factors as a single requirement, and then remove
                # these from each of the other requirements in the list
                remaining_cells = set(c for gp in requirement for c in gp.pos) - set(
                    c for gp in factors for c in gp.pos
                )
                for factor in factors:
                    res.append((factor,))
                res.append(
                    tuple(
                        gp.get_gridded_perm_in_cells(remaining_cells)
                        for gp in requirement
                    )
                )
        return res

    # if a gridded perm G in a requirment has a factor that appears in EVERY gridded
    # perm of some other requirement, then that factor can be removed from G
    # [subrequirement inferral]
    def cleaned_requirements(
        self, requirements: List[Requirement]
    ) -> List[Requirement]:
        """
        Remove the factors of a gridded permutation in a requirement if it is
        implied by another requirement.
        """
        for idx, requirement in enumerate(requirements):
            newgps: List[GriddedPerm] = []
            for gp in requirement:
                cells: List[Cell] = []
                for f in gp.factors():
                    if not self._griddedperm_implied_by_some_requirement(
                        f,
                        chain(
                            islice(requirements, idx),
                            islice(requirements, idx + 1, None),
                        ),
                    ):
                        cells.extend(f.pos)
                newgp = gp.get_gridded_perm_in_cells(cells)
                newgps.append(newgp)
            requirements[idx] = tuple(newgps)
        return requirements

    # We can remove any gridded perm in any requirement if the gridded perm contains
    # an obstruction
    # [requirement component deletion]
    def remove_avoided(
        self,
        requirements: List[Requirement],
        obstructions: Optional[Tuple[GriddedPerm, ...]] = None,
    ) -> List[Requirement]:
        if obstructions is None:
            obstructions = self._obstructions
        res: List[Requirement] = []
        for requirement in requirements:
            # If any gridded permutation in list is empty then you vacuously
            # contain this requirement
            if not all(requirement):
                continue
            cleanreq = tuple(
                gp
                for gp in GriddedPermReduction._minimize(requirement)
                if gp.avoids(*obstructions)
            )
            # If cleanreq is empty, then can not contain this requirement so
            # the tiling is empty.
            if not cleanreq:
                return [tuple()]
            if not (len(cleanreq) == 1 and not cleanreq[0]):
                res.append(cleanreq)
        return res

    @staticmethod
    def _griddedperm_implied_by_requirement(
        griddedperm: GriddedPerm, requirement: Iterable[GriddedPerm]
    ):
        """
        Return True, if the containment of the gridded perm is implied by
        the containment of the requirement.
        """
        return all(griddedperm in gp for gp in requirement)

    # detect if there exists a requirement such that every component in that requirement
    # contains griddedperm
    def _griddedperm_implied_by_some_requirement(
        self,
        griddedperm: GriddedPerm,
        requirements: Optional[Iterable[Requirement]] = None,
    ):
        """
        Return True if the containiment some requirement implies the
        containment griddedperm.
        """
        if requirements is None:
            requirements = self._requirements
        return any(
            GriddedPermReduction._griddedperm_implied_by_requirement(
                griddedperm, requirement
            )
            for requirement in requirements
        )

    def _requirement_implied_by_some_requirement(
        self,
        requirement: Requirement,
        requirements: Iterable[Requirement],
    ) -> bool:
        """
        Return True if one of the requirements implies the containment of requirement.
        """
        return any(
            self._requirement_implied_by_requirement(requirement, req)
            for req in requirements
        )

    @staticmethod
    def _requirement_implied_by_requirement(
        requirement: Requirement, other_requirement: Requirement
    ) -> bool:
        """
        Return True if the containment of other implies a containment of requirement.
        """
        return all(
            any(other_gp.contains(gp) for gp in requirement)
            for other_gp in other_requirement
        )

    @staticmethod
    def _minimize(griddedperms: Iterable[GriddedPerm]) -> Tuple[GriddedPerm, ...]:
        """
        Removes non-minimal gridded permutations from the set.
        """
        perms_by_size: Dict[int, List[GriddedPerm]] = defaultdict(list)
        for gp in griddedperms:
            perms_by_size[len(gp)].append(gp)
        sizes = sorted(perms_by_size.keys())
        if not sizes:
            return tuple()
        minimal_perms = set(perms_by_size[sizes[0]])
        for size in sizes[1:]:
            next_layer = set()
            for gp in perms_by_size[size]:
                if gp.avoids(*minimal_perms):
                    next_layer.add(gp)
            minimal_perms |= next_layer
        return tuple(minimal_perms)

    @staticmethod
    def factors(griddedperms: Tuple[GriddedPerm, ...]) -> Set[GriddedPerm]:
        res: Set[GriddedPerm] = set(griddedperms[0].factors())
        for gp in griddedperms[1:]:
            if not res:
                break
            res = res.intersection(gp.factors())
        return res



from collections import Counter
from functools import reduce
from itertools import chain, combinations
from operator import mul
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from sympy import Eq, Function

from comb_spec_searcher import (
    CartesianProduct,
    CartesianProductStrategy,
    StrategyFactory,
)
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from permuta import Perm
from tilings import GriddedPerm, Tiling
from tilings.algorithms import (
    Factor,
    FactorWithInterleaving,
    FactorWithMonotoneInterleaving,
)
from tilings.assumptions import TrackingAssumption
from tilings.exception import InvalidOperationError
from tilings.misc import multinomial, partitions_iterator
from tilings.strategies.assumption_insertion import (
    AddAssumptionsConstructor,
    AddAssumptionsStrategy,
)

Cell = Tuple[int, int]

__all__ = (
    "FactorFactory",
    "FactorStrategy",
    "FactorWithInterleavingStrategy",
    "FactorWithMonotoneInterleavingStrategy",
)

TempGP = Tuple[
    Tuple[
        Tuple[Union[float, int], ...], Tuple[Union[float, int], ...], Tuple[Cell, ...]
    ],
    ...,
]


class FactorStrategy(CartesianProductStrategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        partition: Iterable[Iterable[Cell]],
        ignore_parent: bool = True,
        workable: bool = True,
    ):
        self.partition = tuple(sorted(tuple(sorted(p)) for p in partition))
        inferrable = any(
            FactorWithInterleavingStrategy.interleaving_rows_and_cols(self.partition)
        )
        super().__init__(
            ignore_parent=ignore_parent, workable=workable, inferrable=inferrable
        )

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        return tuple(comb_class.sub_tiling(cells) for cells in self.partition)

    def extra_parameters(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Strategy does not apply")
        extra_parameters: Tuple[Dict[str, str], ...] = tuple({} for _ in children)
        for parent_var, assumption in zip(
            comb_class.extra_parameters, comb_class.assumptions
        ):
            for idx, child in enumerate(children):
                new_assumption = child.forward_map.map_assumption(assumption).avoiding(
                    child.obstructions
                )
                if new_assumption.gps:
                    child_var = child.get_assumption_parameter(new_assumption)
                    extra_parameters[idx][parent_var] = child_var
        return extra_parameters

    def formal_step(self) -> str:
        """
        Return a string that describe the operation performed on the tiling.
        """
        partition_str = " / ".join(
            f"{{{', '.join(map(str, part))}}}" for part in self.partition
        )
        return f"factor with partition {partition_str}"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        gps_to_combine = tuple(
            tiling.backward_map.map_gp(cast(GriddedPerm, gp))
            for gp, tiling in zip(objs, children)
        )
        temp = [
            ((cell[0], idx), (cell[1], val))
            for gp in gps_to_combine
            for (idx, val), cell in zip(enumerate(gp.patt), gp.pos)
        ]
        temp.sort()
        new_pos = [(idx[0], val[0]) for idx, val in temp]
        new_patt = Perm.to_standard(val for _, val in temp)
        yield GriddedPerm(new_patt, new_pos)

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[GriddedPerm, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        return tuple(
            tiling.forward_map.map_gp(obj.get_gridded_perm_in_cells(part))
            for tiling, part in zip(children, self.partition)
        )

    def __str__(self) -> str:
        return "factor"

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"partition={self.partition}",
                f"ignore_parent={self.ignore_parent}",
                f"workable={self.workable}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    # JSON methods

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the strategy."""
        d: dict = super().to_jsonable()
        d.pop("inferrable")
        d.pop("possibly_empty")
        d["partition"] = self.partition
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FactorStrategy":
        partition = cast(
            Tuple[Tuple[Cell]],
            tuple(tuple(tuple(c) for c in p) for p in d.pop("partition")),
        )
        return cls(partition=partition, **d)


# The following functions are used to determine assumptions needed to count the
# interleavings of a factor. They are also used by AddInterleavingAssumptionStrategy.


class Interleaving(CartesianProduct[Tiling, GriddedPerm]):
    def __init__(
        self,
        parent: Tiling,
        children: Iterable[Tiling],
        extra_parameters: Tuple[Dict[str, str], ...],
        interleaving_parameters: Iterable[Tuple[str, ...]],
        insertion_constructor: Optional[AddAssumptionsConstructor],
    ):
        super().__init__(parent, children, extra_parameters)
        self.interleaving_parameters = tuple(interleaving_parameters)
        self.interleaving_indices = tuple(
            tuple(parent.extra_parameters.index(k) for k in parameters)
            for parameters in interleaving_parameters
        )
        self.insertion_constructor = insertion_constructor

    @staticmethod
    def is_equivalence(is_empty: Optional[Callable[[Tiling], bool]] = None) -> bool:
        return False

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        non_interleaved_terms = super().get_terms(parent_terms, subterms, n)
        interleaved_terms: Terms = Counter()
        for parameters, value in non_interleaved_terms.items():
            # multinomial counts the number of ways to interleave the values k1, ...,kn.
            multiplier = reduce(
                mul,
                [
                    multinomial([parameters[k] for k in int_parameters])
                    for int_parameters in self.interleaving_indices
                ],
                1,
            )
            interleaved_terms[parameters] += multiplier * value
        if self.insertion_constructor:
            new_terms: Terms = Counter()
            for param, value in interleaved_terms.items():
                new_terms[self.insertion_constructor.child_param_map(param)] += value
            return new_terms
        return interleaved_terms

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        for param, objs in super().get_sub_objects(subobjs, n):
            if self.insertion_constructor:
                param = self.insertion_constructor.child_param_map(param)
            yield param, objs

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError


class FactorWithInterleavingStrategy(FactorStrategy):
    def __init__(
        self,
        partition: Iterable[Iterable[Cell]],
        ignore_parent: bool = True,
        workable: bool = True,
        tracked: bool = True,
    ):
        super().__init__(partition, ignore_parent, workable)
        self.tracked = tracked
        self.cols, self.rows = self.interleaving_rows_and_cols(self.partition)

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["tracked"] = self.tracked
        return d

    def __repr__(self) -> str:
        args = ", ".join(
            [
                f"partition={self.partition}",
                f"ignore_parent={self.ignore_parent}",
                f"workable={self.workable}",
                f"tracked={self.tracked}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def is_two_way(self, comb_class: Tiling) -> bool:  # pylint: disable=W0221
        return self.is_reversible(comb_class)

    def is_reversible(self, comb_class: Tiling) -> bool:  # pylint: disable=W0221
        return not bool(self.assumptions_to_add(comb_class))

    def formal_step(self) -> str:
        return "interleaving " + super().formal_step()

    def assumptions_to_add(self, comb_class: Tiling) -> Tuple[TrackingAssumption, ...]:
        """Return the set of assumptions that need to be added to"""
        cols, rows = self.interleaving_rows_and_cols(self.partition)
        return tuple(
            ass
            for ass in chain.from_iterable(
                self._assumptions_to_add(cells, cols, rows) for cells in self.partition
            )
            if ass not in comb_class.assumptions
        )

    @staticmethod
    def interleaving_rows_and_cols(
        partition: Tuple[Tuple[Cell, ...], ...]
    ) -> Tuple[Set[int], Set[int]]:
        """
        Return the set of cols and the set of rows that are being interleaved when
        factoring with partition.
        """
        cols: Set[int] = set()
        rows: Set[int] = set()
        x_seen: Set[int] = set()
        y_seen: Set[int] = set()
        for part in partition:
            cols.update(x for x, _ in part if x in x_seen)
            rows.update(y for _, y in part if y in y_seen)
            x_seen.update(x for x, _ in part)
            y_seen.update(y for _, y in part)
        return cols, rows

    @staticmethod
    def _assumptions_to_add(
        cells: Tuple[Cell, ...], cols: Set[int], rows: Set[int]
    ) -> Tuple[TrackingAssumption, ...]:
        """
        Return the assumptions that should be tracked in the set of cells if we are
        interleaving the given rows and cols.
        """
        col_assumptions = [
            TrackingAssumption(
                [GriddedPerm.point_perm(cell) for cell in cells if x == cell[0]]
            )
            for x in cols
        ]
        row_assumptions = [
            TrackingAssumption(
                [GriddedPerm.point_perm(cell) for cell in cells if y == cell[1]]
            )
            for y in rows
        ]
        return tuple(ass for ass in chain(col_assumptions, row_assumptions) if ass.gps)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        if self.tracked:
            comb_class = comb_class.add_assumptions(self.assumptions_to_add(comb_class))
        return super().decomposition_function(comb_class)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Interleaving:
        if children is None:
            children = self.decomposition_function(comb_class)
        assumptions = self.assumptions_to_add(comb_class)
        insertion_constructor = None
        if assumptions:
            insertion_constructor = AddAssumptionsStrategy(assumptions).constructor(
                comb_class
            )
            comb_class = comb_class.add_assumptions(assumptions)
        interleaving_parameters = self.interleaving_parameters(comb_class)
        if interleaving_parameters and not self.tracked:
            raise NotImplementedError("The interleaving factor was not tracked.")
        return Interleaving(
            comb_class,
            children,
            self.extra_parameters(comb_class, children),
            interleaving_parameters,
            insertion_constructor,
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ):
        raise NotImplementedError

    def interleaving_parameters(self, comb_class: Tiling) -> List[Tuple[str, ...]]:
        """
        Return the parameters on the parent tiling that needed to be interleaved.
        """
        res: List[Tuple[str, ...]] = []
        cols, rows = self.interleaving_rows_and_cols(self.partition)
        for x in cols:
            assumptions = [
                TrackingAssumption(
                    GriddedPerm.point_perm(cell) for cell in cells if x == cell[0]
                )
                for cells in self.partition
            ]
            res.append(
                tuple(
                    comb_class.get_assumption_parameter(ass)
                    for ass in assumptions
                    if ass.gps
                )
            )
        for y in rows:
            assumptions = [
                TrackingAssumption(
                    GriddedPerm.point_perm(cell) for cell in cells if y == cell[1]
                )
                for cells in self.partition
            ]
            res.append(
                tuple(
                    comb_class.get_assumption_parameter(ass)
                    for ass in assumptions
                    if ass.gps
                )
            )
        return res

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        gps_to_combine = tuple(
            tiling.backward_map.map_gp(cast(GriddedPerm, gp))
            for gp, tiling in zip(objs, children)
        )
        all_gps_to_combine: List[TempGP] = [
            tuple(
                (tuple(range(len(gp))), tuple(gp.patt), gp.pos) for gp in gps_to_combine
            )
        ]
        for row in self.rows:
            all_gps_to_combine = self._interleave_row(all_gps_to_combine, row)
        for col in self.cols:
            all_gps_to_combine = self._interleave_col(all_gps_to_combine, col)

        for interleaved_gps_to_combine in all_gps_to_combine:
            temp = [
                ((cell[0], idx), (cell[1], val))
                for gp in interleaved_gps_to_combine
                for idx, val, cell in zip(*gp)
            ]
            temp.sort()
            new_pos = [(idx[0], val[0]) for idx, val in temp]
            new_patt = Perm.to_standard(val for _, val in temp)
            assert not GriddedPerm(new_patt, new_pos).contradictory()
            yield GriddedPerm(new_patt, new_pos)

    def _interleave_row(
        self,
        all_gps_to_combine: List[TempGP],
        row: int,
    ) -> List[TempGP]:
        # pylint: disable=too-many-locals
        res: List[TempGP] = []
        for gps_to_combine in all_gps_to_combine:
            row_points = tuple(
                tuple(
                    (idx, values[idx])
                    for idx, cell in enumerate(position)
                    if cell[1] == row
                )
                for _, values, position in gps_to_combine
            )
            total = sum(len(points) for points in row_points)
            if total == 0:
                res.append(gps_to_combine)
                continue
            min_val = min(val for _, val in chain(*row_points))
            max_val = max(val for _, val in chain(*row_points)) + 1
            temp_values = tuple(
                min_val + i * (max_val - min_val) / total for i in range(total)
            )
            for partition in self._partitions(
                set(temp_values), tuple(len(indices) for indices in row_points)
            ):
                new_gps_to_combine = []
                for part, (indices, values, position), points in zip(
                    partition, gps_to_combine, row_points
                ):
                    new_values = list(values)
                    actual_indices = [
                        idx for _, idx in sorted((val, idx) for idx, val in points)
                    ]
                    for idx, val in zip(actual_indices, sorted(part)):
                        new_values[idx] = val
                    new_gps_to_combine.append((indices, tuple(new_values), position))
                res.append(tuple(new_gps_to_combine))
        return res

    def _interleave_col(
        self,
        all_gps_to_combine: List[TempGP],
        col: int,
    ):
        # pylint: disable=too-many-locals
        res: List[TempGP] = []
        for gps_to_combine in all_gps_to_combine:
            col_points = tuple(
                tuple(
                    (idx, indices[idx])
                    for idx, cell in enumerate(position)
                    if cell[0] == col
                )
                for indices, _, position in gps_to_combine
            )
            total = sum(len(points) for points in col_points)
            if total == 0:
                res.append(gps_to_combine)
                continue
            mindex = min(val for _, val in chain(*col_points))
            maxdex = max(val for _, val in chain(*col_points)) + 1
            temp_indices = tuple(
                mindex + i * (maxdex - mindex) / total for i in range(total)
            )
            for partition in self._partitions(
                set(temp_indices), tuple(len(indices) for indices in col_points)
            ):
                new_gps_to_combine = []
                for part, (indices, values, position), points in zip(
                    partition, gps_to_combine, col_points
                ):
                    new_indices = list(indices)
                    for idx, new_idx in zip([idx for idx, _ in points], sorted(part)):
                        new_indices[idx] = new_idx
                    new_gps_to_combine.append((tuple(new_indices), values, position))
                res.append(tuple(new_gps_to_combine))
        return res

    @staticmethod
    def _partitions(
        values: Set[float], size_of_parts: Tuple[int, ...]
    ) -> Iterator[Tuple[Tuple[float, ...], ...]]:
        if not size_of_parts:
            if not values:
                yield tuple()
            return
        size = size_of_parts[0]
        for part in combinations(values, size):
            for rest in FactorWithInterleavingStrategy._partitions(
                values - set(part), size_of_parts[1:]
            ):
                yield (part,) + rest

    @staticmethod
    def get_eq_symbol() -> str:
        return "="

    @staticmethod
    def get_op_symbol() -> str:
        return "*"


class FactorWithMonotoneInterleavingStrategy(FactorWithInterleavingStrategy):
    pass


class FactorFactory(StrategyFactory[Tiling]):

    FACTOR_ALGO_AND_CLASS = {
        None: (Factor, FactorStrategy),
        "monotone": (
            FactorWithMonotoneInterleaving,
            FactorWithMonotoneInterleavingStrategy,
        ),
        "all": (FactorWithInterleaving, FactorWithInterleavingStrategy),
    }

    def __init__(
        self,
        interleaving: Optional[str] = None,
        unions: bool = False,
        ignore_parent: bool = True,
        workable: bool = True,
        tracked: bool = False,
    ) -> None:
        try:
            self.factor_algo, self.factor_class = self.FACTOR_ALGO_AND_CLASS[
                interleaving
            ]
        except KeyError as e:
            raise InvalidOperationError(
                "Invalid interleaving option. Must be in "
                f"{FactorFactory.FACTOR_ALGO_AND_CLASS.keys()}, "
                f"used {interleaving}"
            ) from e
        self.unions = unions
        self.ignore_parent = ignore_parent
        self.workable = workable
        self.tracked = tracked and self.factor_class in (
            FactorWithInterleavingStrategy,
            FactorWithMonotoneInterleaving,
        )

    def __call__(self, comb_class: Tiling) -> Iterator[FactorStrategy]:
        factor_algo = self.factor_algo(comb_class)
        if factor_algo.factorable():
            min_comp = tuple(tuple(part) for part in factor_algo.get_components())
            if self.unions:
                for partition in partitions_iterator(min_comp):
                    components = tuple(
                        tuple(chain.from_iterable(part)) for part in partition
                    )
                    yield self._build_strategy(components, workable=False)
            yield self._build_strategy(min_comp, workable=self.workable)

    def _build_strategy(
        self, components: Tuple[Tuple[Cell, ...], ...], workable: bool
    ) -> FactorStrategy:
        """
        Build the factor strategy for the given components.

        It ensure that a plain factor rule is returned.
        """
        interleaving = any(
            FactorWithInterleavingStrategy.interleaving_rows_and_cols(components)
        )
        if interleaving:
            # pylint: disable=E1123
            return cast(
                FactorWithInterleavingStrategy,
                self.factor_class(
                    components,
                    ignore_parent=self.ignore_parent,
                    workable=workable,
                    tracked=self.tracked,
                ),
            )
        return FactorStrategy(
            components, ignore_parent=self.ignore_parent, workable=workable
        )

    def __str__(self) -> str:
        if self.factor_class is FactorStrategy:
            s = "factor"
        elif self.factor_class is FactorWithInterleavingStrategy:
            s = "factor with interleaving"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            s = "factor with monotone interleaving"
        else:
            raise Exception("Invalid interleaving type")
        if self.unions:
            s = "unions of " + s
        if self.tracked:
            s = "tracked " + s
        return s

    def __repr__(self) -> str:
        if self.factor_class is FactorStrategy:
            interleaving = None
        elif self.factor_class is FactorWithInterleavingStrategy:
            interleaving = "all"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            interleaving = "monotone"
        else:
            raise ValueError("Invalid interleaving type")
        args = ", ".join(
            [
                f"interleaving={interleaving!r}",
                f"unions={self.unions}",
                f"ignore_parent={self.ignore_parent}",
                f"workable={self.workable}",
                f"tracked={self.tracked}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    def to_jsonable(self) -> dict:
        d: dict = super().to_jsonable()
        if self.factor_class is FactorStrategy:
            interleaving = None
        elif self.factor_class is FactorWithInterleavingStrategy:
            interleaving = "all"
        elif self.factor_class is FactorWithMonotoneInterleavingStrategy:
            interleaving = "monotone"
        else:
            raise ValueError("Invalid interleaving type")

        d["interleaving"] = interleaving
        d["unions"] = self.unions
        d["ignore_parent"] = self.ignore_parent
        d["workable"] = self.workable
        d["tracked"] = self.tracked
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "FactorFactory":
        return cls(**d)



import abc
from importlib import import_module
from itertools import chain
from typing import TYPE_CHECKING, Iterable, List, Optional, Tuple, Type

from permuta import Perm

from .griddedperm import GriddedPerm

Cell = Tuple[int, int]

if TYPE_CHECKING:
    from tilings import Tiling


class TrackingAssumption:
    """
    An assumption used to keep track of the occurrences of a set of gridded
    permutations.
    """

    def __init__(self, gps: Iterable[GriddedPerm]):
        self.gps = tuple(sorted(set(gps)))

    @classmethod
    def from_cells(cls, cells: Iterable[Cell]) -> "TrackingAssumption":
        gps = [GriddedPerm.single_cell((0,), cell) for cell in cells]
        return cls(gps)

    def get_cells(self) -> Tuple[Cell, ...]:
        assert all(len(gp) == 1 for gp in self.gps)
        return tuple(gp.pos[0] for gp in self.gps)

    def avoiding(
        self,
        obstructions: Iterable[GriddedPerm],
        active_cells: Optional[Iterable[Cell]] = None,
    ) -> "TrackingAssumption":
        """
        Return the tracking absumption where all of the gridded perms avoiding
        the obstructions are removed. If active_cells is not None, then any
        assumptions involving a cell not in active_cells will be removed.
        """
        obstructions = tuple(obstructions)
        if active_cells is not None:
            return self.__class__(
                tuple(
                    gp
                    for gp in self.gps
                    if all(cell in active_cells for cell in gp.pos)
                    and gp.avoids(*obstructions)
                )
            )
        return self.__class__(tuple(gp for gp in self.gps if gp.avoids(*obstructions)))

    def get_value(self, gp: GriddedPerm) -> int:
        """
        Return the number of occurrences of each of the gridded perms being track in
        the gridded perm gp.
        """
        return len(list(chain.from_iterable(p.occurrences_in(gp) for p in self.gps)))

    def get_components(self, tiling: "Tiling") -> List[List[GriddedPerm]]:
        """
        Return the lists of gps that count exactly one occurrence.
        Only implemented for when a size one gp is in a point cell.
        """
        return [
            [gp] for gp in self.gps if len(gp) == 1 and gp.pos[0] in tiling.point_cells
        ]

    def remove_components(self, tiling: "Tiling") -> "TrackingAssumption":
        """
        Return the TrackingAssumption found by removing all the components
        found by the get_components method.
        """
        gps_to_remove = set(chain.from_iterable(self.get_components(tiling)))
        return self.__class__(gp for gp in self.gps if gp not in gps_to_remove)

    def to_jsonable(self) -> dict:
        """Return a dictionary form of the assumption."""
        c = self.__class__
        return {
            "class_module": c.__module__,
            "assumption": c.__name__,
            "gps": [gp.to_jsonable() for gp in self.gps],
        }

    @classmethod
    def from_dict(cls, d: dict) -> "TrackingAssumption":
        """Return the assumption from the json dict representation."""
        module = import_module(d["class_module"])
        AssClass: Type["TrackingAssumption"] = getattr(module, d["assumption"])
        assert issubclass(
            AssClass, TrackingAssumption
        ), "Not a valid TrackingAssumption"
        gps = [GriddedPerm.from_dict(gp) for gp in d["gps"]]
        return AssClass(gps)

    def __eq__(self, other) -> bool:
        if other.__class__ == TrackingAssumption:
            return bool(self.gps == other.gps)
        return NotImplemented

    def __lt__(self, other) -> bool:
        if isinstance(other, TrackingAssumption):
            key_self = (self.__class__.__name__, self.gps)
            key_other = (other.__class__.__name__, other.gps)
            return key_self < key_other
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.gps)

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.gps})"

    def __str__(self):
        if all(len(gp) == 1 for gp in self.gps):
            cells = ", ".join(str(gp.pos[0]) for gp in self.gps)
            return f"can count points in cell{'s' if len(self.gps) > 1 else ''} {cells}"
        return f"can count occurrences of {', '.join(str(gp) for gp in self.gps)}"


class ComponentAssumption(TrackingAssumption):
    """
    An assumption used to keep track of the number of components in a
    region of a tiling.

    In order to inherit from TrackingAssumption, the set of cells should be
    given as a set of length 1 gridded perms using each cell. This ensures
    most strategies work without change.
    """

    def __init__(self, gps: Iterable[GriddedPerm]):
        super().__init__(gps)
        assert all(len(gp) == 1 for gp in self.gps)
        self.cells = frozenset(gp.pos[0] for gp in self.gps)

    @abc.abstractmethod
    def decomposition(self, perm: Perm) -> List[Perm]:
        """Count the number of component in a permutation."""

    @abc.abstractmethod
    def tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        """Return the components of a given tiling."""

    @abc.abstractmethod
    def opposite_tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        """Return the alternative components of a given tiling."""

    @abc.abstractmethod
    def one_or_fewer_components(self, tiling: "Tiling", cell: Cell) -> bool:
        """Return True if the cell contains one or fewer components."""

    def get_components(self, tiling: "Tiling") -> List[List[GriddedPerm]]:
        sub_tiling = tiling.sub_tiling(self.cells)
        separated_tiling, fwd_map = sub_tiling.row_and_column_separation_with_mapping()
        back_map = {b: a for a, b in fwd_map.items()}
        components = self.tiling_decomposition(separated_tiling)
        return [
            [
                GriddedPerm.point_perm(sub_tiling.backward_map.map_cell(back_map[cell]))
                for cell in comp
            ]
            for comp in components
            if self.is_component(comp, separated_tiling)
        ]

    def is_component(self, cells: List[Cell], tiling: "Tiling") -> bool:
        """
        Return True if cells form a component on the tiling.

        Cells are assumed to have come from cell_decomposition.
        """
        sub_tiling = tiling.sub_tiling(cells)
        skew_cells = self.opposite_tiling_decomposition(sub_tiling)

        if any(
            scells[0] in sub_tiling.positive_cells
            and self.one_or_fewer_components(sub_tiling, scells[0])
            for scells in skew_cells
            if len(scells) == 1
        ):
            return True

        def is_positive(scells) -> bool:
            return any(
                all(any(cell in gp.pos for cell in scells) for gp in req)
                for req in sub_tiling.requirements
            )

        return sum(1 for scells in skew_cells if is_positive(scells)) > 1

    def cell_decomposition(self, tiling: "Tiling"):
        sub_tiling = tiling.sub_tiling(self.cells)
        return [
            [sub_tiling.backward_map.map_cell(cell) for cell in comp]
            for comp in self.tiling_decomposition(sub_tiling)
        ]

    def get_value(self, gp: GriddedPerm) -> int:
        """
        Return the number of components in the tracked region of the gridded perm.
        """
        subgp = gp.get_gridded_perm_in_cells(self.cells)
        return len(self.decomposition(subgp.patt))

    def __eq__(self, other) -> bool:
        if isinstance(other, ComponentAssumption) and self.__class__ == other.__class__:
            return bool(self.gps == other.gps)
        return NotImplemented

    def __repr__(self) -> str:
        return self.__class__.__name__ + f"({self.gps})"

    def __str__(self):
        return f"can count components in cells {self.cells}"

    def __hash__(self) -> int:
        return hash(self.gps)


class SumComponentAssumption(ComponentAssumption):
    def decomposition(self, perm: Perm) -> List[Perm]:
        return perm.sum_decomposition()  # type: ignore

    def tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        return tiling.sum_decomposition()

    def opposite_tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        return tiling.skew_decomposition()

    def one_or_fewer_components(self, tiling: "Tiling", cell: Cell) -> bool:
        return GriddedPerm.single_cell(Perm((0, 1)), cell) in tiling.obstructions

    def __str__(self):
        return f"can count sum components in cells {self.cells}"

    def __hash__(self) -> int:
        return hash(self.gps)


class SkewComponentAssumption(ComponentAssumption):
    def decomposition(self, perm: Perm) -> List[Perm]:
        return perm.skew_decomposition()  # type: ignore

    def tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        return tiling.skew_decomposition()

    def opposite_tiling_decomposition(self, tiling: "Tiling") -> List[List[Cell]]:
        return tiling.sum_decomposition()

    def one_or_fewer_components(self, tiling: "Tiling", cell: Cell) -> bool:
        return GriddedPerm.single_cell(Perm((1, 0)), cell) in tiling.obstructions

    def __str__(self):
        return f"can count skew components in cells {self.cells}"

    def __hash__(self) -> int:
        return hash(self.gps)



from collections import defaultdict
from functools import reduce
from itertools import product
from operator import mul
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple

from sympy import Eq, Function, var

from comb_spec_searcher import Constructor, Strategy
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.typing import (
    Parameters,
    RelianceProfile,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from comb_spec_searcher.utils import compositions
from tilings import GriddedPerm, Tiling
from tilings.algorithms import factor
from tilings.assumptions import (
    SkewComponentAssumption,
    SumComponentAssumption,
    TrackingAssumption,
)

Cell = Tuple[int, int]


class Split(Constructor):
    """
    The constructor used to count when a variable is counted by some multiple
    disjoint subvariables.
    """

    def __init__(self, split_parameters: Dict[str, Tuple[str, ...]]):
        self.split_parameters = split_parameters

    def get_equation(self, lhs_func: Function, rhs_funcs: Tuple[Function, ...]) -> Eq:
        rhs_func = rhs_funcs[0]
        subs: Dict[Any, List[Any]] = defaultdict(list)
        for parent, children in self.split_parameters.items():
            for child in children:
                subs[var(child)].append(var(parent))
        var_subs = {
            parent: reduce(mul, children, 1) for parent, children in subs.items()
        }
        return Eq(lhs_func, rhs_func.subs(var_subs, simultaneous=True))

    def reliance_profile(self, n: int, **parameters: int) -> RelianceProfile:
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        raise NotImplementedError

    def _valid_compositions(self, **parameters: int) -> Iterator[Dict[str, int]]:
        """
        For each parameter which splits according to split_parameters, it will
        take a composition of the value for parameter and assign it to the
        subparameters it is split into.

        TODO: this should consider reliance profiles, and when variables are
        sub variable of other variables.
        """

        def compositions_dict(value: int, parameters: Tuple[str, ...]):
            for comp in compositions(
                value,
                len(parameters),
                (0,) * len(parameters),
                (None,) * len(parameters),
            ):
                yield dict(zip(parameters, comp))

        def union_params(
            sub_params: Tuple[Dict[str, int], ...]
        ) -> Optional[Dict[str, int]]:
            new_params: Dict[str, int] = {}
            for params in sub_params:
                for k, val in params.items():
                    if k in new_params:
                        if val != new_params[k]:
                            return None
                    else:
                        new_params[k] = val
            return new_params

        for sub_params in product(
            *[
                compositions_dict(val, self.split_parameters[key])
                for key, val in parameters.items()
            ]
        ):
            new_params = union_params(sub_params)
            if new_params is not None:
                yield new_params

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[Tuple[Parameters, Tuple[List[Optional[GriddedPerm]], ...]]]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ):
        raise NotImplementedError

    def equiv(
        self, other: "Constructor", data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError("Required for bijections")


class SplittingStrategy(Strategy[Tiling, GriddedPerm]):
    """
    A strategy which splits each TrackingAssumption into sub TrackingAssumptions,
    according to the factors of the underlying tiling.

    TODO: iterate over all possible union of factors
    """

    FACTOR_ALGO = {
        "none": factor.Factor,
        "monotone": factor.FactorWithMonotoneInterleaving,
        "all": factor.FactorWithInterleaving,
    }

    def __init__(
        self,
        interleaving: str = "none",
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
    ):
        try:
            self.factor_class = SplittingStrategy.FACTOR_ALGO[interleaving]
        except KeyError as e:
            raise ValueError(
                "interleaving argument must be in "
                f"{list(SplittingStrategy.FACTOR_ALGO)}"
            ) from e
        super().__init__(
            ignore_parent=ignore_parent,
            inferrable=inferrable,
            possibly_empty=possibly_empty,
            workable=workable,
        )

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling):
        return False

    def is_reversible(self, comb_class: Tiling):
        return False

    def shifts(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]]
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply
        return (0,)

    def decomposition_function(self, comb_class: Tiling) -> Optional[Tuple[Tiling]]:
        if not comb_class.assumptions:
            return None
        components = self.factor_class(comb_class.remove_assumptions()).get_components()
        if len(components) == 1:
            return None
        new_assumptions: List[TrackingAssumption] = []
        for ass in comb_class.assumptions:
            new_assumptions.extend(self._split_assumption(ass, components))
        new_tiling = Tiling(
            comb_class.obstructions,
            comb_class.requirements,
            sorted(new_assumptions),
            remove_empty_rows_and_cols=False,
            derive_empty=False,
            simplify=False,
            sorted_input=True,
        )
        new_tiling.clean_assumptions()
        return (new_tiling,)

    def _split_assumption(
        self, assumption: TrackingAssumption, components: Tuple[Set[Cell], ...]
    ) -> List[TrackingAssumption]:
        if isinstance(assumption, SkewComponentAssumption):
            return self._split_skew_assumption(assumption)
        if isinstance(assumption, SumComponentAssumption):
            return self._split_sum_assumption(assumption)
        return self._split_tracking_assumption(assumption, components)

    @staticmethod
    def _split_tracking_assumption(
        assumption: TrackingAssumption, components: Tuple[Set[Cell], ...]
    ) -> List[TrackingAssumption]:
        split_gps: List[List[GriddedPerm]] = [[] for _ in range(len(components))]
        for gp in assumption.gps:
            for idx, component in enumerate(components):
                if all(cell in component for cell in gp.pos):
                    split_gps[idx].append(gp)
                    # only add to one list
                    break
            else:
                # gridded perm can't be partitioned, so the partition can't be
                # partitioned
                return [assumption]
        return [assumption.__class__(gps) for gps in split_gps if gps]

    def _split_skew_assumption(
        self, assumption: SkewComponentAssumption
    ) -> List[TrackingAssumption]:
        decomposition = self.skew_decomposition(assumption.cells)
        return [
            SkewComponentAssumption(
                GriddedPerm.single_cell((0,), cell) for cell in cells
            )
            for cells in decomposition
        ]

    def _split_sum_assumption(
        self, assumption: SumComponentAssumption
    ) -> List[TrackingAssumption]:
        decomposition = self.sum_decomposition(assumption.cells)
        return [
            SumComponentAssumption(
                GriddedPerm.single_cell((0,), cell) for cell in cells
            )
            for cells in decomposition
        ]

    @staticmethod
    def sum_decomposition(
        cells: Iterable[Cell], skew: bool = False
    ) -> List[List[Cell]]:
        """
        Returns the sum decomposition of the cells.
        If skew is True then returns the skew decomposition instead.
        """
        cells = sorted(cells)
        decomposition: List[List[Cell]] = []
        while len(cells) > 0:
            x = cells[0][0]  # x boundary, maximum in both cases
            y = cells[0][1]  # y boundary, maximum in sum, minimum in skew
            change = True
            while change:
                change = False
                for c in cells:
                    if c[0] <= x:
                        if (skew and c[1] < y) or (not skew and c[1] > y):
                            y = c[1]
                            change = True
                    if (skew and c[1] >= y) or (not skew and c[1] <= y):
                        if c[0] > x:
                            x = c[0]
                            change = True
            decomposition.append([])
            new_cells = []
            for c in cells:
                if c[0] <= x:
                    decomposition[-1].append(c)
                else:
                    new_cells.append(c)
            cells = new_cells
        return decomposition

    def skew_decomposition(self, cells: Iterable[Cell]) -> List[List[Cell]]:
        """
        Returns the skew decomposition of the cells
        """
        return self.sum_decomposition(cells, skew=True)

    def constructor(
        self, comb_class: Tiling, children: Optional[Tuple[Tiling, ...]] = None
    ) -> Split:
        if children is None:
            children = self.decomposition_function(comb_class)
            if children is None:
                raise StrategyDoesNotApply("Can't split the tracking assumption")
        child = children[0]
        split_parameters: Dict[str, Tuple[str, ...]] = {"n": ("n",)}
        components = self.factor_class(comb_class.remove_assumptions()).get_components()
        for idx, assumption in enumerate(comb_class.assumptions):
            split_assumptions = self._split_assumption(assumption, components)
            child_vars = tuple(
                sorted(f"k_{child.assumptions.index(ass)}" for ass in split_assumptions)
            )
            split_parameters[f"k_{idx}"] = child_vars
        return Split(split_parameters)

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Constructor:
        raise NotImplementedError

    def formal_step(self) -> str:
        return "splitting the assumptions"

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        """
        The forward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        """
        The backward direction of the underlying bijection used for object
        generation and sampling.
        """
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["interleaving"] = next(
            k
            for k, v in SplittingStrategy.FACTOR_ALGO.items()
            if v == self.factor_class
        )
        return d

    def __repr__(self) -> str:
        if self.factor_class is factor.Factor:
            interleaving = "none"
        elif self.factor_class is factor.FactorWithInterleaving:
            interleaving = "all"
        elif self.factor_class is factor.FactorWithMonotoneInterleaving:
            interleaving = "monotone"
        else:
            raise ValueError("Unknown factor class")
        args = ", ".join(
            [
                f"interleaving={interleaving!r}",
                f"ignore_parent={self.ignore_parent}",
                f"inferrable={self.inferrable}",
                f"possibly_empty={self.possibly_empty}",
                f"workable={self.workable}",
            ]
        )
        return f"{self.__class__.__name__}({args})"

    @classmethod
    def from_dict(cls, d: dict) -> "SplittingStrategy":
        return cls(**d)

    @staticmethod
    def get_eq_symbol() -> str:
        return "↣"



import re
import subprocess
import sys
from importlib.util import find_spec
from typing import Optional, Tuple

try:
    if tuple(
        map(lambda z: int(re.sub("[^0-9]", "", z)), sys.version.split()[0].split("."))
    ) < (3, 8, 0):
        from importlib_metadata import version  # pylint: disable=import-error
    else:
        from importlib.metadata import version  # pylint: disable=ungrouped-imports
except ImportError:

    def _fake_version(_str: str) -> str:

        print("=== If you are using older version of python than 3.8 ===")
        print(f"To use for {sys.executable} you need importlib-metadata.")
        print(f"{sys.executable} -m pip install importlib-metadata", flush=True)
        return "0.0.0"

    version = _fake_version


_MIN_VERSION = (0, 2, 1)
_NOT_FOUND, _OLD_VERSION, _SUCCESS = range(3)


def in_path_version(interpreter: str) -> Optional[Tuple[int, ...]]:
    try:
        return next(
            (
                tuple(map(int, line.split(" ")[1].split(".")))
                for line in subprocess.run(
                    [
                        interpreter,
                        "-m",
                        "pip",
                        "show",
                        "tilingsgui",
                    ],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
                .stdout.decode("utf-8")
                .splitlines()
                if line.startswith("Version")
            ),
            None,
        )
    except (FileNotFoundError, subprocess.CalledProcessError):
        return None


def search_path_interpreter(interpreter) -> Tuple[int, str]:
    vers = in_path_version(interpreter)
    if vers is None:
        return _NOT_FOUND, interpreter
    if vers < _MIN_VERSION:
        return _OLD_VERSION, interpreter
    return _SUCCESS, interpreter


def search_current_interpreter() -> Tuple[int, str]:
    spec = find_spec("tilingsgui")
    if spec is None or spec.loader is None:
        return _NOT_FOUND, sys.executable
    if tuple(map(int, version("tilingsgui").split("."))) < _MIN_VERSION:
        return _OLD_VERSION, sys.executable
    return _SUCCESS, sys.executable


def look_for_tilingsgui() -> Tuple[int, str]:
    best_status = -1
    for search in (
        search_current_interpreter,
        lambda: search_path_interpreter("python"),
        lambda: search_path_interpreter("python3"),
        lambda: search_path_interpreter("pypy3"),
        lambda: search_path_interpreter("python3.8"),
        lambda: search_path_interpreter("python3.7"),
        lambda: search_path_interpreter("python3.6"),
        lambda: search_path_interpreter("python3.9"),
    ):
        status, interpreter = search()
        best_status = max(best_status, status)
        if best_status == _SUCCESS:
            break
    return best_status, interpreter


def run_gui(tiling_json: str) -> None:
    status, interpreter = look_for_tilingsgui()
    if status == _SUCCESS:
        subprocess.run(
            [interpreter, "-m", "tilingsgui.main", "-j", tiling_json], check=True
        )
    elif status == _OLD_VERSION:
        print("Found tilingsgui but version is outdated.")
        print("your_python_interpreter -m pip install tilingsgui --upgrade", flush=True)
    else:
        print("Did not find tilingsgui.")
        print("your_python_interpreter -m pip install tilingsgui", flush=True)



from typing import Dict, Iterator, Optional, Tuple

from comb_spec_searcher import Strategy, StrategyFactory
from comb_spec_searcher.exception import StrategyDoesNotApply
from comb_spec_searcher.strategies import Rule
from tilings import GriddedPerm, Tiling
from tilings.strategies.dummy_constructor import DummyConstructor


class RelaxAssumptionStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(self, child: Tiling, assumption_idx: int, **kwargs):
        self.child = child
        self.assumption_idx = assumption_idx
        super().__init__(**kwargs)

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return False

    def is_reversible(self, comb_class: Tiling) -> bool:
        return False

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        return (0,)

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling]:
        try:
            assumption = self.child.assumptions[self.assumption_idx]
        except IndexError as e:
            raise StrategyDoesNotApply from e
        parent = self.child.add_obstructions(assumption.gps)
        if parent != comb_class:
            raise StrategyDoesNotApply
        return (self.child,)

    def formal_step(self) -> str:
        return f"the assumption at index {self.assumption_idx} is relaxed"

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DummyConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DummyConstructor()

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DummyConstructor:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DummyConstructor()

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def extra_parameters(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["child"] = self.child.to_jsonable()
        d["assumption_idx"] = self.assumption_idx
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "RelaxAssumptionStrategy":
        return cls(Tiling.from_dict(d.pop("child")), d.pop("assumption_idx"), **d)


class RelaxAssumptionFactory(StrategyFactory[Tiling]):
    def __call__(self, comb_class: Tiling) -> Iterator[Rule]:
        for idx, assumption in enumerate(comb_class.assumptions):
            parent = comb_class.add_obstructions(assumption.gps)
            yield RelaxAssumptionStrategy(comb_class, idx)(parent, (comb_class,))

    @classmethod
    def from_dict(cls, d: dict) -> "RelaxAssumptionFactory":
        return cls(**d)

    def __str__(self) -> str:
        return "Relax assumption"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"



from collections import Counter
from itertools import chain
from typing import Callable, Dict, Iterator, List, Optional, Tuple

import sympy

from comb_spec_searcher import Strategy
from comb_spec_searcher.strategies import Constructor
from comb_spec_searcher.strategies.constructor import Complement, DisjointUnion
from comb_spec_searcher.strategies.strategy import StrategyFactory
from comb_spec_searcher.typing import (
    CombinatorialClassType,
    CombinatorialObjectType,
    Parameters,
    SubObjects,
    SubRecs,
    SubSamplers,
    SubTerms,
    Terms,
)
from tilings import GriddedPerm, Tiling
from tilings.algorithms import Fusion


class DivideByN(DisjointUnion[CombinatorialClassType, CombinatorialObjectType]):
    """
    A constructor that works as disjoint union
    but divides the values by n + shift.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        shift: int,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.shift = shift
        self.initial_conditions = {
            n: parent.get_terms(n) for n in range(1 - self.shift)
        }
        super().__init__(parent, children, extra_parameters)

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        # TODO: d/dx [ x**shift * lhsfun ] / x**(shift - 1) = A + B + ...
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        if n + self.shift <= 0:
            return self.initial_conditions[n]
        terms = super().get_terms(parent_terms, subterms, n)
        return Counter({key: value // (n + self.shift) for key, value in terms.items()})

    def get_sub_objects(
        self, subobjs: SubObjects, n: int
    ) -> Iterator[
        Tuple[Parameters, Tuple[List[Optional[CombinatorialObjectType]], ...]]
    ]:
        raise NotImplementedError

    def random_sample_sub_objects(
        self,
        parent_count: int,
        subsamplers: SubSamplers,
        subrecs: SubRecs,
        n: int,
        **parameters: int,
    ) -> Tuple[Optional[CombinatorialObjectType], ...]:
        raise NotImplementedError

    @staticmethod
    def get_eq_symbol() -> str:
        return "?"

    def __str__(self):
        return "divide by n"

    def equiv(
        self, other: Constructor, data: Optional[object] = None
    ) -> Tuple[bool, Optional[object]]:
        raise NotImplementedError


class ReverseDivideByN(Complement[CombinatorialClassType, CombinatorialObjectType]):
    """
    The complement version of DivideByN.
    It works as Complement, but multiplies by n + shift the original left hand side.
    """

    def __init__(
        self,
        parent: CombinatorialClassType,
        children: Tuple[CombinatorialClassType, ...],
        idx: int,
        shift: int,
        extra_parameters: Optional[Tuple[Dict[str, str], ...]] = None,
    ):
        self.shift = shift
        self.initial_conditions = {
            n: children[idx].get_terms(n) for n in range(1 - self.shift)
        }
        super().__init__(parent, children, idx, extra_parameters)

    def get_equation(
        self, lhs_func: sympy.Function, rhs_funcs: Tuple[sympy.Function, ...]
    ) -> sympy.Eq:
        # TODO: rhs_funcs[0] should be a derivative etc, see DivideByN.get_equation.
        raise NotImplementedError

    def get_terms(
        self, parent_terms: Callable[[int], Terms], subterms: SubTerms, n: int
    ) -> Terms:
        if n + self.shift <= 0:
            return self.initial_conditions[n]
        parent_terms_mapped: Terms = Counter()
        for param, value in subterms[0](n).items():
            if value:
                # This is the only change from complement
                N = n + self.shift
                assert N > 0
                parent_terms_mapped[self._parent_param_map(param)] += value * N
        children_terms = subterms[1:]
        for child_terms, param_map in zip(children_terms, self._children_param_maps):
            # we subtract from total
            for param, value in child_terms(n).items():
                mapped_param = self._parent_param_map(param_map(param))
                parent_terms_mapped[mapped_param] -= value
                assert parent_terms_mapped[mapped_param] >= 0
                if parent_terms_mapped[mapped_param] == 0:
                    parent_terms_mapped.pop(mapped_param)

        return parent_terms_mapped

    def __str__(self):
        return "reverse divide by n"


class UnfusionColumnStrategy(Strategy[Tiling, GriddedPerm]):
    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
        cols: bool = True,
    ):
        self.cols = cols
        super().__init__(ignore_parent, inferrable, possibly_empty, workable)

    def can_be_equivalent(self) -> bool:
        return False

    def is_two_way(self, comb_class: Tiling) -> bool:
        return True

    def is_reversible(self, comb_class: Tiling) -> bool:
        return True

    def shifts(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]],
    ) -> Tuple[int, ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        return tuple(0 for _ in range(self.width(comb_class)))

    def width(self, comb_class: Tiling) -> int:
        if self.cols:
            return comb_class.dimensions[0]
        return comb_class.dimensions[1]

    def decomposition_function(self, comb_class: Tiling) -> Tuple[Tiling, ...]:
        res = []
        for idx in range(self.width(comb_class)):
            if self.cols:
                algo = Fusion(comb_class, col_idx=idx)
            else:
                algo = Fusion(comb_class, row_idx=idx)
            obs = chain(
                *[algo.unfuse_gridded_perm(ob) for ob in comb_class.obstructions]
            )
            reqs = [
                [gp for req_gp in req_list for gp in algo.unfuse_gridded_perm(req_gp)]
                for req_list in comb_class.requirements
            ]
            ass = [
                ass.__class__(
                    [
                        gp
                        for ass_gp in ass.gps
                        for gp in algo.unfuse_gridded_perm(ass_gp)
                    ]
                )
                for ass in comb_class.assumptions
            ]
            res.append(Tiling(obs, reqs, ass))
        return tuple(res)

    def formal_step(self) -> str:
        if self.cols:
            return "unfuse columns strategy"
        return "unfuse rows strategy"

    def constructor(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> DivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return DivideByN(
            comb_class,
            children,
            len(children),
            self.extra_parameters(comb_class, children),
        )

    def reverse_constructor(
        self,
        idx: int,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> ReverseDivideByN:
        if children is None:
            children = self.decomposition_function(comb_class)
        return ReverseDivideByN(
            comb_class,
            children,
            idx,
            len(children),
            self.extra_parameters(comb_class, children),
        )

    def backward_map(
        self,
        comb_class: Tiling,
        objs: Tuple[Optional[GriddedPerm], ...],
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Iterator[GriddedPerm]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def forward_map(
        self,
        comb_class: Tiling,
        obj: GriddedPerm,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Optional[GriddedPerm], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        raise NotImplementedError

    def extra_parameters(
        self,
        comb_class: Tiling,
        children: Optional[Tuple[Tiling, ...]] = None,
    ) -> Tuple[Dict[str, str], ...]:
        if children is None:
            children = self.decomposition_function(comb_class)
        res = []
        for idx in range(self.width(comb_class)):
            if self.cols:
                algo = Fusion(comb_class, col_idx=idx)
            else:
                algo = Fusion(comb_class, row_idx=idx)
            params: Dict[str, str] = {}
            for ass in comb_class.assumptions:
                mapped_ass = ass.__class__(
                    [
                        children[idx].forward_map.map_gp(gp)
                        for ass_gp in ass.gps
                        for gp in algo.unfuse_gridded_perm(ass_gp)
                    ]
                )
                params[comb_class.get_assumption_parameter(ass)] = children[
                    idx
                ].get_assumption_parameter(mapped_ass)
            res.append(params)
        return tuple(res)

    @classmethod
    def from_dict(cls, d: dict) -> "UnfusionColumnStrategy":
        return cls(**d)


class UnfusionRowStrategy(UnfusionColumnStrategy):
    def __init__(
        self,
        ignore_parent: bool = False,
        inferrable: bool = True,
        possibly_empty: bool = True,
        workable: bool = True,
        cols: bool = False,
    ):
        super().__init__(ignore_parent, inferrable, possibly_empty, workable, cols)


class UnfusionFactory(StrategyFactory[Tiling]):
    def __init__(self, max_width: int = 4, max_height: int = 4) -> None:
        self.max_width = max_width
        self.max_height = max_height
        super().__init__()

    def __call__(self, comb_class: Tiling) -> Iterator[UnfusionColumnStrategy]:
        width, height = comb_class.dimensions
        if width <= self.max_width:
            yield UnfusionColumnStrategy()
        if height <= self.max_height:
            yield UnfusionRowStrategy()

    def __str__(self) -> str:
        return "unfusion strategy"

    def __repr__(self) -> str:
        return (
            self.__class__.__name__
            + f"(max_width={self.max_width}, max_height={self.max_height})"
        )

    def to_jsonable(self) -> dict:
        d = super().to_jsonable()
        d["max_width"] = self.max_width
        d["max_height"] = self.max_height
        return d

    @classmethod
    def from_dict(cls, d: dict) -> "UnfusionFactory":
        return cls(max_width=d["max_width"], max_height=d["max_height"])



from collections import defaultdict, deque
from functools import partial
from typing import Callable, DefaultDict, Deque, Dict, Optional, Set, Tuple

from comb_spec_searcher.bijection import EqPathParallelSpecFinder, SpecMap
from comb_spec_searcher.comb_spec_searcher import CombinatorialSpecificationSearcher
from comb_spec_searcher.isomorphism import Bijection, Isomorphism
from comb_spec_searcher.specification import CombinatorialSpecification
from comb_spec_searcher.specification_extrator import RulePathToAtomExtractor
from comb_spec_searcher.strategies.rule import AbstractRule, ReverseRule, Rule
from comb_spec_searcher.typing import Terms
from tilings import GriddedPerm, Tiling
from tilings.assumptions import TrackingAssumption
from tilings.strategies import BasicVerificationStrategy
from tilings.strategies.fusion.fusion import FusionRule
from tilings.tilescope import TileScope

AssumptionLabels = Dict[int, Set[TrackingAssumption]]


class _TermCacher:
    """A cache wrapper for getting terms."""

    MAX_TERMS_LEN = 5

    def __init__(self) -> None:
        self._cache: DefaultDict[Tiling, Dict[int, Terms]] = defaultdict(dict)

    def get_term_function_from_tiling(self, tiling: Tiling) -> Callable[[int], Terms]:
        """Return a cached version of `get_terms` given a tiling."""
        return partial(self._tiling_f, tiling)

    def get_term_function_from_rule(self, rule: FusionRule) -> Callable[[int], Terms]:
        """Return a cached version of `get_terms` given a fusion rule."""
        return partial(self._rule_f, rule)

    def _tiling_f(self, tiling: Tiling, n: int) -> Terms:
        return self._f(tiling, tiling.get_terms, n)

    def _rule_f(self, rule: FusionRule, n: int) -> Terms:
        return self._f(rule.comb_class, rule.get_terms, n)

    def _f(self, cache_key: Tiling, func: Callable[[int], Terms], n: int) -> Terms:
        cached = self._cache[cache_key]
        retval = cached.get(n, None)
        if retval is None:
            retval = func(n)
            cached[n] = retval
        return retval


class _AssumptionPathTracker:
    """Tracks assumptions down to atoms with labels."""

    def __init__(
        self,
        path1: Deque[Tuple[AbstractRule[Tiling, GriddedPerm], int]],
        path2: Deque[Tuple[AbstractRule[Tiling, GriddedPerm], int]],
    ) -> None:
        if 0 in (len(path1), len(path2)):
            raise ValueError("Paths should contain at least one rule.")

        # Number of non-equivalence rules should be the same
        assert sum(1 for r, _ in path1 if not r.is_equivalence()) == sum(
            1 for r, _ in path2 if not r.is_equivalence()
        )

        self.nxt_label = 0
        self.path1 = path1
        self.path2 = path2

        # rules and indices set to that of root and first child in each path
        self.r1, self.idx1 = _AssumptionPathTracker._get_next(path1)
        self.r2, self.idx2 = _AssumptionPathTracker._get_next(path2)

        # If root contains any assumptions, they are gathered here
        # These are a map of labels (ids) to assumptions
        self.assumptions1, self.assumptions2 = self._init_assumptions()

    def _init_assumptions(self) -> Tuple[AssumptionLabels, AssumptionLabels]:
        a1: AssumptionLabels = {}
        a2: AssumptionLabels = {}
        # If any label at root, we store them at the first label
        if self.r1.comb_class.assumptions:
            a1[self.nxt_label] = set(self.r1.comb_class.assumptions)
        if self.r2.comb_class.assumptions:
            a2[self.nxt_label] = set(self.r2.comb_class.assumptions)
        # Increment next label if root had any assumptions
        self.nxt_label = max(len(a1), len(a2))
        return a1, a2

    @staticmethod
    def _get_next(
        path: Deque[Tuple[AbstractRule, int]]
    ) -> Tuple[Rule[Tiling, GriddedPerm], int]:
        """Get next rule and the index of its child we traverse through."""
        r, idx = path.popleft()
        assert isinstance(r, Rule)
        assert not path or r.children[idx] == path[0][0].comb_class
        return r, idx

    @staticmethod
    def _get_par_map(rule: Rule[Tiling, GriddedPerm], idx: int) -> Dict[str, str]:
        """Given a rule and an index of its child, give the extra parameters map from
        the parent to that child.
        """
        if not isinstance(rule, ReverseRule):
            return rule.strategy.extra_parameters(rule.comb_class, rule.children)[idx]
        # If the rule is a reverse rule, we assume that it has only one children
        # as that is the only used reverse rule that have been implemented so far.
        assert len(rule.children) == 1
        pmap = rule.strategy.extra_parameters(rule.children[0], (rule.comb_class,))[idx]
        assert len(pmap) == len(set(pmap.values()))  # invertible
        return {v: k for k, v in pmap.items()}

    @staticmethod
    def _forward_all_labels(
        assumptions: AssumptionLabels,
        rule: Rule[Tiling, GriddedPerm],
        idx: int,
    ) -> AssumptionLabels:
        """Recreate the dictionary of labels to assumptions by mapping each assumption
        to the corresponding assumption in the idx-th child of rule, if there is any.
        If the label's value becomes empty, we remove it.
        """
        d: AssumptionLabels = {}
        parmap = _AssumptionPathTracker._get_par_map(rule, idx)
        for i, s in assumptions.items():
            new_set = set()
            for assumption in s:
                parent_var = rule.comb_class.get_assumption_parameter(assumption)
                # If parent's assumption is in child
                if parent_var in parmap:
                    child_var = parmap[parent_var]
                    child_assumption = rule.children[idx].get_assumption(child_var)
                    new_set.add(child_assumption)
            if new_set:
                d[i] = new_set
        return d

    def _step_through_equivalences(self):
        """
        Map assumptions though equivalences assuming they won't introduce new ones.
        """
        while self.r1.is_equivalence():
            self.assumptions1 = _AssumptionPathTracker._forward_all_labels(
                self.assumptions1, self.r1, self.idx1
            )
            self.r1, self.idx1 = _AssumptionPathTracker._get_next(self.path1)
        while self.r2.is_equivalence():
            self.assumptions2 = _AssumptionPathTracker._forward_all_labels(
                self.assumptions2, self.r2, self.idx2
            )
            self.r2, self.idx2 = _AssumptionPathTracker._get_next(self.path2)

    def _step_forward(self) -> None:
        """
        A step forward will perform the map for one non-equivalence rule for each path.
        """
        self._step_through_equivalences()
        self.assumptions1 = _AssumptionPathTracker._forward_all_labels(
            self.assumptions1, self.r1, self.idx1
        )
        self.assumptions2 = _AssumptionPathTracker._forward_all_labels(
            self.assumptions2, self.r2, self.idx2
        )

    @staticmethod
    def _find_new_assumptions(
        rule: Rule[Tiling, GriddedPerm],
        assumptions: AssumptionLabels,
    ) -> Set[TrackingAssumption]:
        """The assumption in the current class that don't correspond to any label."""
        return set(
            assumption
            for assumption in rule.comb_class.assumptions
            if all(assumption not in v for v in assumptions.values())
        )

    def _check_for_new_assumptions(self) -> None:
        a1 = _AssumptionPathTracker._find_new_assumptions(self.r1, self.assumptions1)
        a2 = _AssumptionPathTracker._find_new_assumptions(self.r2, self.assumptions2)

        # Add any new atoms to the label tra
        if a1:
            self.assumptions1[self.nxt_label] = a1
        if a2:
            self.assumptions2[self.nxt_label] = a2

        # Increment the label system if new ones
        if a1 or a2:
            self.nxt_label += 1

    def assumptions_match_down_to_atom(self) -> bool:
        while True:
            # Map assumptions to next node
            self._step_forward()

            # If there aren't any more rules, we are done as we should be at an atom now
            if not self.path1 and not self.path2:
                break

            # Assign next rules
            self.r1, self.idx1 = _AssumptionPathTracker._get_next(self.path1)
            self.r2, self.idx2 = _AssumptionPathTracker._get_next(self.path2)

            # Search for new assumptions that will recieve a new label
            self._check_for_new_assumptions()

        # When at the atoms, the labels should match
        return sorted(self.assumptions1.keys()) == sorted(self.assumptions2.keys())


class FusionParallelSpecFinder(
    EqPathParallelSpecFinder[Tiling, GriddedPerm, Tiling, GriddedPerm]
):
    """A specialized parallel specification finder for fusion."""

    def __init__(self, searcher1: TileScope, searcher2: TileScope):
        super().__init__(searcher1, searcher2)
        self._term_cacher = _TermCacher()

    def _pre_expand(
        self,
        searcher1: CombinatorialSpecificationSearcher[Tiling],
        searcher2: CombinatorialSpecificationSearcher[Tiling],
    ) -> None:
        # pylint: disable=no-self-use
        searcher1.strategy_pack = searcher1.strategy_pack.add_verification(
            BasicVerificationStrategy(), replace=True
        )
        searcher2.strategy_pack = searcher2.strategy_pack.add_verification(
            BasicVerificationStrategy(), replace=True
        )

    def _atom_path_match(self, id1: int, id2: int, sp1: SpecMap, sp2: SpecMap) -> bool:
        path1, path2 = self._get_paths(sp1, sp2)
        return _AssumptionPathTracker(path1, path2).assumptions_match_down_to_atom()

    def _rule_match(
        self,
        rule1: AbstractRule[Tiling, GriddedPerm],
        rule2: AbstractRule[Tiling, GriddedPerm],
    ) -> bool:
        # pylint: disable=no-self-use
        if not isinstance(rule1, Rule):
            return not isinstance(rule2, Rule)
        if not isinstance(rule2, Rule):
            return False
        # If comparing two fusion rules, terms are required
        if isinstance(rule1, FusionRule) and isinstance(rule2, FusionRule):
            terms = (
                self._term_cacher.get_term_function_from_tiling(rule1.comb_class),
                self._term_cacher.get_term_function_from_tiling(rule2.comb_class),
                _TermCacher.MAX_TERMS_LEN,
            )
            return rule1.constructor.equiv(rule2.constructor, terms)[0]
        return rule1.constructor.equiv(rule2.constructor)[0]

    def _get_paths(
        self, sp1: SpecMap, sp2: SpecMap
    ) -> Tuple[Deque[Tuple[AbstractRule, int]], Deque[Tuple[AbstractRule, int]]]:
        path1 = RulePathToAtomExtractor(
            self._pi1.searcher.start_label,
            self._pi1.root_eq_label,
            FusionParallelSpecFinder._create_tree(sp1, self._pi1.root_eq_label),
            self._pi1.ruledb,
            self._pi1.searcher.classdb,
            [(label, idx) for label, _, idx, _ in reversed(self._path)],
        ).rule_path()
        path2 = RulePathToAtomExtractor(
            self._pi2.searcher.start_label,
            self._pi2.root_eq_label,
            FusionParallelSpecFinder._create_tree(sp2, self._pi2.root_eq_label),
            self._pi2.ruledb,
            self._pi2.searcher.classdb,
            [(label, idx) for _, label, _, idx in reversed(self._path)],
        ).rule_path()
        return path1, path2


class FusionIsomorphism(Isomorphism[Tiling, GriddedPerm, Tiling, GriddedPerm]):
    """Does additional checks for atoms and fusion constructors."""

    def __init__(
        self,
        spec1: CombinatorialSpecification[Tiling, GriddedPerm],
        spec2: CombinatorialSpecification[Tiling, GriddedPerm],
        term_cacher: Optional[_TermCacher] = None,
    ) -> None:
        self._term_cacher = term_cacher if term_cacher else _TermCacher()
        super().__init__(spec1, spec2)

    def _atom_match(
        self,
        atom1: Tiling,
        atom2: Tiling,
        rule1: AbstractRule[Tiling, GriddedPerm],
        rule2: AbstractRule[Tiling, GriddedPerm],
    ) -> bool:
        """Returns true if atoms match, false otherwise."""
        # pylint: disable=no-self-use
        if not super()._atom_match(atom1, atom2, rule1, rule2):
            return False
        return self._assumption_match_down_to_atom()

    def _assumption_match_down_to_atom(self) -> bool:
        path1: Deque[Tuple[AbstractRule[Tiling, GriddedPerm], int]] = deque([])
        path2: Deque[Tuple[AbstractRule[Tiling, GriddedPerm], int]] = deque([])
        curr1, curr2 = self.root1, self.root2
        # Collect the rules and child indices down to our current atom
        for step1, step2 in self._path_tracker:
            # spec 1
            for idx in step1:
                rule = self._rules1[curr1]
                i = tuple(j for j, c in enumerate(rule.children) if not c.is_empty())[
                    idx
                ]
                path1.append((rule, i))
                curr1 = rule.children[i]
            # spec 2
            for idx in step2:
                rule = self._rules2[curr2]
                i = tuple(j for j, c in enumerate(rule.children) if not c.is_empty())[
                    idx
                ]
                path2.append((rule, i))
                curr2 = rule.children[i]
        return _AssumptionPathTracker(path1, path2).assumptions_match_down_to_atom()

    def _constructor_match(
        self,
        rule1: Rule[Tiling, GriddedPerm],
        rule2: Rule[Tiling, GriddedPerm],
        curr1: Tiling,
        curr2: Tiling,
    ) -> bool:
        if not (isinstance(rule1, FusionRule) and isinstance(rule2, FusionRule)):
            return super()._constructor_match(rule1, rule2, curr1, curr2)
        terms = (
            self._term_cacher.get_term_function_from_rule(rule1),
            self._term_cacher.get_term_function_from_rule(rule2),
            _TermCacher.MAX_TERMS_LEN,
        )
        are_eq, data = rule1.constructor.equiv(rule2.constructor, terms)
        if not are_eq:
            return False
        assert isinstance(data, list) and 0 < len(data) < 3
        # If data contains two values, we can't determine the direction of the match
        if (len(data) == 1 and data[0]) or (
            len(data) == 2 and self._fusion_reverse_match(rule1, rule2)
        ):
            self._index_data[(curr1, curr2)] = True  # Reverse match cached
        return True

    def _fusion_reverse_match(self, rule1: FusionRule, rule2: FusionRule) -> bool:
        # Find fused parameter in parents by going downwards
        parent_param1 = FusionIsomorphism._assumption_bfs_to_parent(rule1, self._rules1)
        parent_param2 = FusionIsomorphism._assumption_bfs_to_parent(rule2, self._rules2)
        # If we fail to find parent in either, we use normal matching
        if parent_param1 is None or parent_param2 is None:
            return False
        # Reverse match if there is a mismatch in where the parameters land
        return (
            parent_param1 in rule1.constructor.left_sided_parameters
            and parent_param2 in rule2.constructor.right_sided_parameters
        ) or (
            parent_param1 in rule1.constructor.right_sided_parameters
            and parent_param2 in rule2.constructor.left_sided_parameters
        )

    @staticmethod
    def _get_par_maps(_rule: Rule[Tiling, GriddedPerm]) -> Tuple[Dict[str, str], ...]:
        if not isinstance(_rule, ReverseRule):
            return _rule.strategy.extra_parameters(_rule.comb_class, _rule.children)
        assert len(_rule.children) == 1  # We currently only have 1-1 rev rules
        return _rule.strategy.extra_parameters(_rule.children[0], (_rule.comb_class,))

    @staticmethod
    def _assumption_bfs_to_parent(
        fusion_rule: FusionRule,
        rules: Dict[Tiling, AbstractRule[Tiling, GriddedPerm]],
    ) -> Optional[str]:
        """A memory enhanced bfs to find parent that tracks parameter."""
        mem: Set[Tiling] = {fusion_rule.children[0]}
        queue: Deque[Tuple[Tiling, str]] = deque(
            [(fusion_rule.children[0], fusion_rule.constructor.fuse_parameter)]
        )
        while queue:
            curr, par = queue.popleft()
            if curr == fusion_rule.comb_class:
                return par
            rule = rules[curr]
            assert isinstance(rule, Rule)
            for i, pmap in enumerate(FusionIsomorphism._get_par_maps(rule)):
                child = rule.children[i]
                if not child.is_atom() and par in pmap and child not in mem:
                    queue.append((child, pmap[par]))
                    mem.add(child)


class FusionBijection(Bijection[Tiling, GriddedPerm, Tiling, GriddedPerm]):
    """Replaces base Isomorphism with FusionIsomorphism."""

    @classmethod
    def construct(
        cls,
        spec: CombinatorialSpecification[Tiling, GriddedPerm],
        other: CombinatorialSpecification[Tiling, GriddedPerm],
    ) -> Optional["FusionBijection"]:
        iso = FusionIsomorphism(spec, other)
        if not iso.are_isomorphic():
            return None
        return cls(spec, other, iso.get_order(), iso.get_order_data())
